import logging
from pathlib import Path

from blends.ast.build_graph import get_ast_graph_from_content
from blends.content.content import get_content_by_path
from blends.ctx import ctx
from blends.models import Content, Graph, NId, SyntaxGraph
from blends.stack.validation import validate_stack_graph_graph
from blends.syntax.cfg.generate import add_syntax_cfg
from blends.syntax.dispatchers import DISPATCHERS_BY_LANG
from blends.syntax.models import ReaderLogicError, SyntaxGraphArgs, SyntaxMetadata
from blends.syntax.readers.common.missing_node import reader as missing_node_reader

LOGGER = logging.getLogger(__name__)


def generic(args: SyntaxGraphArgs) -> NId:
    node_type = args.ast_graph.nodes[args.n_id]["label_type"]
    lang_dispatchers = DISPATCHERS_BY_LANG[args.language]
    if dispatcher := lang_dispatchers.get(node_type):
        try:
            return dispatcher(args)
        except ReaderLogicError as e:
            LOGGER.debug("Reader logic error for %s in %s: %s", node_type, args.language.name, e)
            return missing_node_reader(args, node_type)

    LOGGER.debug("Missing syntax reader for %s in %s", node_type, args.language.name)

    return missing_node_reader(args, node_type)


def get_syntax_graph(
    ast_graph: Graph, content: Content, *, with_cfg: bool = True, with_metadata: bool = False
) -> SyntaxGraph | None:
    syntax_graph = SyntaxGraph()

    *_, last_n_id = ast_graph.nodes
    max_ast_id = int(last_n_id)

    syntax_metadata = SyntaxMetadata(
        class_path=[],
        scope_stack=[NId(1)] if ctx.has_feature_flag("StackGraph") else [],
        next_synthetic_id=max_ast_id + 1,
    )
    if with_metadata:
        syntax_graph.add_node(
            NId(0),
            label_type="Metadata",
            structure={},
            instances={},
            imports=[],
            path=str(content.path),
        )

    generic(
        SyntaxGraphArgs(
            generic,
            str(content.path),
            content.language,
            ast_graph,
            syntax_graph,
            NId(1),
            syntax_metadata,
        ),
    )

    if ctx.has_feature_flag("StackGraph"):
        errors = validate_stack_graph_graph(syntax_graph)
        if errors:
            max_errors = 20
            head = errors[:max_errors]
            extra = len(errors) - len(head)
            details = "; ".join(head)
            if extra > 0:
                details = f"{details}; and {extra} more"
            LOGGER.warning("Stack graph validation errors in %s: %s", content.path, details)

    if with_cfg:
        syntax_graph = add_syntax_cfg(syntax_graph, is_multifile=with_metadata)

    return syntax_graph


def get_syntax_graph_from_path(
    path: Path, *, with_cfg: bool = True, with_metadata: bool = False
) -> SyntaxGraph | None:
    content = get_content_by_path(path)
    if content is None:
        return None

    ast_graph = get_ast_graph_from_content(content)
    if ast_graph is None:
        return None

    return get_syntax_graph(ast_graph, content, with_cfg=with_cfg, with_metadata=with_metadata)



from dataclasses import dataclass
from typing import Optional

_OFFSET_MUST_BE_NONNEGATIVE = "offset must be nonnegative"


@dataclass(frozen=True, slots=True)
class SymbolStackVariable:
    var_id: int

    @classmethod
    def new(cls, var_id: int) -> Optional["SymbolStackVariable"]:
        if var_id < 1:
            return None
        return cls(var_id=var_id)

    @classmethod
    def initial(cls) -> "SymbolStackVariable":
        return cls(var_id=1)

    def with_offset(self, offset: int) -> "SymbolStackVariable":
        if offset < 0:
            raise ValueError(_OFFSET_MUST_BE_NONNEGATIVE)
        return SymbolStackVariable(var_id=self.var_id + offset)

    def as_int(self) -> int:
        return self.var_id


@dataclass(frozen=True, slots=True)
class ScopeStackVariable:
    var_id: int

    @classmethod
    def new(cls, var_id: int) -> Optional["ScopeStackVariable"]:
        if var_id < 1:
            return None
        return cls(var_id=var_id)

    @classmethod
    def initial(cls) -> "ScopeStackVariable":
        return cls(var_id=1)

    def with_offset(self, offset: int) -> "ScopeStackVariable":
        if offset < 0:
            raise ValueError(_OFFSET_MUST_BE_NONNEGATIVE)
        return ScopeStackVariable(var_id=self.var_id + offset)

    def as_int(self) -> int:
        return self.var_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    c_ids = adj_ast(args.ast_graph, args.n_id)
    equalizer = match_ast_d(args.ast_graph, args.n_id, "=")
    min_children_for_variable_declaration = 3

    if len(c_ids) >= min_children_for_variable_declaration and equalizer:
        var_name = node_to_str(args.ast_graph, c_ids[0])
        return build_variable_declaration_node(args, var_name, None, c_ids[2])

    ignored_types = {"="}
    filtered_ids = [
        _id for _id in c_ids if args.ast_graph.nodes[_id]["label_type"] not in ignored_types
    ]

    return build_expression_statement_node(args, iter(filtered_ids))



ACCESS_MODIFIER = "access_modifier"
ACCESS_MODIFIERS = "access_modifiers"
ANNOTATION_ID = "annotation_id"
ARGUMENT_NAME = "argument_name"
ARGUMENTS_ID = "arguments_id"
ATTRIBUTES_ID = "attributes_id"
BLOCK_ID = "block_id"
CASE_EXPRESSION = "case_expression"
CATCH_DECLARATION = "catch_declaration"
COMMENT = "comment"
CONDITION_ID = "condition_id"
CONSTRUCTOR_ID = "constructor_id"
DECLARATION_ID = "declaration_id"
EXPRESSION = "expression"
EXPRESSION_ID = "expression_id"
FALSE_ID = "false_id"
IMPORT_TYPE = "import_type"
IMPORTS = "imports"
INHERITED_CLASS = "inherited_class"
INSTANCES = "instances"
INITIALIZER = "initializer"
INITIALIZER_ID = "initializer_id"
INITIALIZERS = "initializers"
ITERABLE_ITEM_ID = "iterable_item_id"
KEY_ID = "key_id"
LABEL_ALIAS = "label_alias"
LABEL_C = "label_c"
LABEL_L = "label_l"
LABEL_TYPE = "label_type"
LEFT_ID = "left_id"
MEMBER = "member"
METHOD_NAME = "method_name"
MODIFIERS_ID = "modifiers_id"
NAME = "name"
NODE_TYPE = "node_type"
OBJECT = "object"
OBJECT_ID = "object_id"
OPERAND_ID = "operand_id"
OPERATOR = "operator"
PARAMETER_MODE = "parameter_mode"
PARAMETERS_ID = "parameters_id"
PATH = "path"
RESOURCES_ID = "resources_id"
RIGHT_ID = "right_id"
SELECTOR_NAME = "selector_name"
STRUCTURE = "structure"
SG_NODE_TYPE = "sg_node_type"
SYMBOL = "symbol"
TF_REFERENCE = "tf_reference"
TRUE_ID = "true_id"
UPDATE_ID = "update_id"
VALUE = "value"
VALUE_ID = "value_id"
VALUE_TYPE = "value_type"
VARIABLE = "variable"
VARIABLE_ID = "variable_id"
VARIABLE_TYPE = "variable_type"

STACK_GRAPH_IS_DEFINITION = "stack_graph_is_definition"
STACK_GRAPH_IS_EXPORTED = "stack_graph_is_exported"
STACK_GRAPH_IS_IMPORT_BINDING = "stack_graph_is_import_binding"
STACK_GRAPH_IMPORT_MODULE_STEM = "stack_graph_import_module_stem"
STACK_GRAPH_IMPORT_SOURCE_SYMBOL = "stack_graph_import_source_symbol"
STACK_GRAPH_IS_REFERENCE = "stack_graph_is_reference"
STACK_GRAPH_KIND = "stack_graph_kind"
STACK_GRAPH_PRECEDENCE = "stack_graph_precedence"
STACK_GRAPH_SCOPE = "stack_graph_scope"
STACK_GRAPH_SCOPE_TYPE = "stack_graph_scope_type"
STACK_GRAPH_SYMBOL = "stack_graph_symbol"

VALID_SYNTAX_ATTRIBUTES: frozenset[str] = frozenset(
    {
        ACCESS_MODIFIER,
        ACCESS_MODIFIERS,
        ANNOTATION_ID,
        ARGUMENT_NAME,
        ARGUMENTS_ID,
        ATTRIBUTES_ID,
        BLOCK_ID,
        CASE_EXPRESSION,
        CATCH_DECLARATION,
        COMMENT,
        CONDITION_ID,
        CONSTRUCTOR_ID,
        DECLARATION_ID,
        EXPRESSION,
        EXPRESSION_ID,
        FALSE_ID,
        IMPORT_TYPE,
        IMPORTS,
        INHERITED_CLASS,
        INITIALIZER,
        INITIALIZER_ID,
        INITIALIZERS,
        INSTANCES,
        ITERABLE_ITEM_ID,
        KEY_ID,
        LABEL_ALIAS,
        LABEL_C,
        LABEL_L,
        LABEL_TYPE,
        LEFT_ID,
        MEMBER,
        METHOD_NAME,
        MODIFIERS_ID,
        NAME,
        NODE_TYPE,
        OBJECT,
        OBJECT_ID,
        OPERAND_ID,
        OPERATOR,
        PARAMETER_MODE,
        PARAMETERS_ID,
        PATH,
        RESOURCES_ID,
        RIGHT_ID,
        SELECTOR_NAME,
        SG_NODE_TYPE,
        STACK_GRAPH_IS_DEFINITION,
        STACK_GRAPH_IS_EXPORTED,
        STACK_GRAPH_IS_IMPORT_BINDING,
        STACK_GRAPH_IMPORT_MODULE_STEM,
        STACK_GRAPH_IMPORT_SOURCE_SYMBOL,
        STACK_GRAPH_IS_REFERENCE,
        STACK_GRAPH_KIND,
        STACK_GRAPH_PRECEDENCE,
        STACK_GRAPH_SCOPE,
        STACK_GRAPH_SCOPE_TYPE,
        STACK_GRAPH_SYMBOL,
        STRUCTURE,
        SYMBOL,
        TF_REFERENCE,
        TRUE_ID,
        UPDATE_ID,
        VALUE,
        VALUE_ID,
        VALUE_TYPE,
        VARIABLE,
        VARIABLE_ID,
        VARIABLE_TYPE,
    }
)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    body_id = n_attrs["label_field_body"]
    if graph.nodes[body_id]["label_type"] == "expression_statement":
        body_id = match_ast(graph, body_id)["__0__"]
    if graph.nodes[body_id]["label_type"] == "parenthesized_expression":
        body_id = match_ast(graph, body_id)["__1__"]

    cond_id = n_attrs["label_field_condition"]
    if graph.nodes[cond_id]["label_type"] == "parenthesized_expression":
        cond_id = match_ast(graph, cond_id)["__1__"]

    return build_while_statement_node(args, body_id, cond_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    attr_id = args.ast_graph.nodes[args.n_id]["label_field_attribute"]
    obj_id = args.ast_graph.nodes[args.n_id]["label_field_object"]
    member = node_to_str(args.ast_graph, attr_id)
    expression = node_to_str(args.ast_graph, obj_id)
    return build_member_access_node(args, member, expression, obj_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.object_creation import (
    build_object_creation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    c_ids = match_ast(graph, args.n_id, "new", "arguments")
    name = ""
    initializer_id = c_ids.get("__0__")
    if initializer_id:
        name = node_to_str(graph, initializer_id)

        if graph.nodes[initializer_id]["label_type"] in {"name", "qualified_name"}:
            initializer_id = None

    arguments_id = c_ids.get("arguments")

    return build_object_creation_node(args, name, arguments_id, initializer_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.else_clause import (
    build_else_clause_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = match_ast(graph, args.n_id, "else")
    body_id = childs.get("__0__")

    if body_id and graph.nodes[body_id]["label_type"] == "expression_statement":
        body_id = match_ast(graph, body_id).get("__0__")

    if body_id and graph.nodes[body_id]["label_type"] == "parenthesized_expression":
        body_id = match_ast(graph, body_id).get("__1__")

    if not body_id:
        body_id = adj_ast(graph, args.n_id)[-1]

    return build_else_clause_node(args, body_id)



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    push_scoped_symbol_node_attributes,
    push_symbol_node_attributes,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_symbol_lookup_node(args: SyntaxGraphArgs, symbol: str) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        symbol=symbol,
        label_type="SymbolLookup",
    )

    if ctx.has_feature_flag("StackGraph"):
        scope_stack = args.metadata.setdefault("scope_stack", [])
        if scope_stack and (current_scope := scope_stack[-1]):
            args.syntax_graph.update_node(
                args.n_id,
                push_scoped_symbol_node_attributes(
                    symbol=symbol,
                    scope=current_scope,
                    precedence=0,
                ),
            )
            add_edge(
                args.syntax_graph,
                Edge(source=args.n_id, sink=current_scope, precedence=0),
            )
        else:
            args.syntax_graph.update_node(
                args.n_id,
                push_symbol_node_attributes(symbol=symbol, precedence=0),
            )

    return args.n_id



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.scala import (
    annotation as scala_annotation,
)
from blends.syntax.readers.scala import (
    argument_list as scala_argument_list,
)
from blends.syntax.readers.scala import (
    array_node as scala_array_node,
)
from blends.syntax.readers.scala import (
    assignment as scala_assignment,
)
from blends.syntax.readers.scala import (
    binary_expression as scala_binary_expression,
)
from blends.syntax.readers.scala import (
    boolean_literal as scala_boolean_literal,
)
from blends.syntax.readers.scala import (
    call_expression as scala_call_expression,
)
from blends.syntax.readers.scala import (
    class_definition as scala_class_definition,
)
from blends.syntax.readers.scala import (
    comment as scala_comment,
)
from blends.syntax.readers.scala import (
    conditional_expression as scala_conditional_expression,
)
from blends.syntax.readers.scala import (
    do_statement as scala_do_statement,
)
from blends.syntax.readers.scala import (
    except_clause as scala_except_clause,
)
from blends.syntax.readers.scala import (
    execution_block as scala_execution_block,
)
from blends.syntax.readers.scala import (
    field_expression as scala_field_expression,
)
from blends.syntax.readers.scala import (
    finally_clause as scala_finally_clause,
)
from blends.syntax.readers.scala import (
    for_statement as scala_for_statement,
)
from blends.syntax.readers.scala import (
    identifier as scala_identifier,
)
from blends.syntax.readers.scala import (
    import_global as scala_import_global,
)
from blends.syntax.readers.scala import (
    interface_declaration as scala_interface_declaration,
)
from blends.syntax.readers.scala import (
    lambda_expression as scala_lambda_expression,
)
from blends.syntax.readers.scala import (
    method_declaration as scala_method_declaration,
)
from blends.syntax.readers.scala import (
    number_literal as scala_number_literal,
)
from blends.syntax.readers.scala import (
    object_creation_expression as scala_object_creation_expression,
)
from blends.syntax.readers.scala import (
    package_declaration as scala_package_declaration,
)
from blends.syntax.readers.scala import (
    parameter as scala_parameter,
)
from blends.syntax.readers.scala import (
    parameter_list as scala_parameter_list,
)
from blends.syntax.readers.scala import (
    parenthesized_expression as scala_parenthesized_expression,
)
from blends.syntax.readers.scala import (
    program as scala_program,
)
from blends.syntax.readers.scala import (
    property_declaration as scala_property_declaration,
)
from blends.syntax.readers.scala import (
    return_statement as scala_return_statement,
)
from blends.syntax.readers.scala import (
    string_literal as scala_string_literal,
)
from blends.syntax.readers.scala import (
    switch_section as scala_switch_section,
)
from blends.syntax.readers.scala import (
    switch_statement as scala_switch_statement,
)
from blends.syntax.readers.scala import (
    throw_statement as scala_throw_statement,
)
from blends.syntax.readers.scala import (
    try_statement as scala_try_statement,
)
from blends.syntax.readers.scala import (
    unary_expression as scala_unary_expression,
)
from blends.syntax.readers.scala import (
    variable_declaration as scala_variable_declaration,
)
from blends.syntax.readers.scala import (
    while_statement as scala_while_statement,
)

SCALA_DISPATCHERS: Dispatcher = {
    "annotation": scala_annotation.reader,
    "ascription_expression": scala_parameter.reader,
    "finally_clause": scala_finally_clause.reader,
    "parameters": scala_parameter_list.reader,
    "arguments": scala_argument_list.reader,
    "function_definition": scala_method_declaration.reader,
    "function_declaration": scala_method_declaration.reader,
    "comment": scala_comment.reader,
    "block_comment": scala_comment.reader,
    "template_body": scala_execution_block.reader,
    "block": scala_execution_block.reader,
    "case_block": scala_execution_block.reader,
    "call_expression": scala_call_expression.reader,
    "class_definition": scala_class_definition.reader,
    "if_expression": scala_conditional_expression.reader,
    "val_definition": scala_variable_declaration.reader,
    "var_definition": scala_variable_declaration.reader,
    "boolean_literal": scala_boolean_literal.reader,
    "integer_literal": scala_number_literal.reader,
    "floating_point_literal": scala_number_literal.reader,
    "for_expression": scala_for_statement.reader,
    "character_literal": scala_string_literal.reader,
    "string": scala_string_literal.reader,
    "interpolated_string_expression": scala_string_literal.reader,
    "identifier": scala_identifier.reader,
    "type_identifier": scala_identifier.reader,
    "object_definition": scala_class_definition.reader,
    "compilation_unit": scala_program.reader,
    "assignment_expression": scala_assignment.reader,
    "while_expression": scala_while_statement.reader,
    "infix_expression": scala_binary_expression.reader,
    "parenthesized_expression": scala_parenthesized_expression.reader,
    "do_while_expression": scala_do_statement.reader,
    "match_expression": scala_switch_statement.reader,
    "parameter": scala_parameter.reader,
    "class_parameter": scala_parameter.reader,
    "catch_clause": scala_except_clause.reader,
    "case_clause": scala_switch_section.reader,
    "try_expression": scala_try_statement.reader,
    "return_expression": scala_return_statement.reader,
    "import_declaration": scala_import_global.reader,
    "lambda_expression": scala_lambda_expression.reader,
    "bindings": scala_parameter_list.reader,
    "binding": scala_parameter.reader,
    "class_parameters": scala_parameter_list.reader,
    "trait_definition": scala_interface_declaration.reader,
    "field_expression": scala_field_expression.reader,
    "instance_expression": scala_object_creation_expression.reader,
    "generic_function": scala_call_expression.reader,
    "throw_expression": scala_throw_statement.reader,
    "type_parameters": scala_argument_list.reader,
    "tuple_expression": scala_array_node.reader,
    "package_clause": scala_package_declaration.reader,
    "type_definition": scala_property_declaration.reader,
    "indented_block": scala_execution_block.reader,
    "wildcard": scala_identifier.reader,
    "postfix_expression": scala_unary_expression.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    children = adj_ast(graph, args.n_id)
    operator = None
    min_children_for_assignment_with_operator = 3
    if len(children) >= min_children_for_assignment_with_operator:
        var_id = children[0]
        operator = node_to_str(graph, children[1])
        val_id = children[2]
    else:
        var_id = children[0]
        val_id = children[-1]

    return build_assignment_node(args, var_id, val_id, operator)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    member_n_id = nodes[args.n_id]["label_field_name"]
    name_n_id = match_ast_d(graph, member_n_id, "name") or member_n_id

    member = nodes[name_n_id].get("label_text", "")

    obj_p_id = nodes[args.n_id].get("label_field_scope")
    expression: str = ""
    if obj_p_id:
        expression = nodes[obj_p_id].get("label_text", "")
        if expression:
            obj_id = obj_p_id
        else:
            obj_c_ids = adj_ast(graph, obj_p_id)
            obj_id = obj_c_ids[0]
            expression = node_to_str(graph, obj_id)

    return build_member_access_node(args, member, expression, obj_id)



import pytest

from blends.models import (
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.partial_path import (
    PartialPath,
    PartialPathResolutionError,
    PartialPathResolutionErrorCode,
    PartialScopedSymbol,
    PartialScopeStack,
    PartialScopeStackBindings,
    PartialSymbolStack,
    PartialSymbolStackBindings,
    ScopeStackVariable,
    SymbolStackVariable,
)
from blends.stack.view import (
    StackGraphView,
)


def _view_for_graph(graph: SyntaxGraph) -> StackGraphView:
    return StackGraphView.from_syntax_graph(graph, path="test.py")


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_stack_variable_with_offset() -> None:
    initial_var_id = 1
    offset_amount = 2
    expected_offset_id = 3
    variable = SymbolStackVariable.initial()
    offset = variable.with_offset(offset_amount)
    assert variable.as_int() == initial_var_id
    assert offset.as_int() == expected_offset_id


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_stack_variable_with_offset() -> None:
    initial_var_id = 1
    offset_amount = 2
    expected_offset_id = 3
    variable = ScopeStackVariable.initial()
    offset = variable.with_offset(offset_amount)
    assert variable.as_int() == initial_var_id
    assert offset.as_int() == expected_offset_id


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_scope_stack_unify_binds_suffix() -> None:
    bindings = PartialScopeStackBindings.new()
    var = ScopeStackVariable.initial()
    lhs = PartialScopeStack(scopes=(1, 2), variable=None)
    rhs = PartialScopeStack(scopes=(1,), variable=var)

    result = lhs.unify(rhs, bindings)

    assert result == lhs
    assert bindings.get(var) == PartialScopeStack(scopes=(2,), variable=None)


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_scope_stack_unify_mismatch_raises() -> None:
    bindings = PartialScopeStackBindings.new()
    lhs = PartialScopeStack(scopes=(1,), variable=None)
    rhs = PartialScopeStack(scopes=(2,), variable=None)

    with pytest.raises(PartialPathResolutionError) as excinfo:
        lhs.unify(rhs, bindings)

    exception = excinfo.value  # type: ignore[attr-defined]
    assert exception.code == PartialPathResolutionErrorCode.SCOPE_STACK_UNSATISFIED


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_symbol_stack_unify_with_attached_scopes() -> None:
    scope_bindings = PartialScopeStackBindings.new()
    symbol_bindings = PartialSymbolStackBindings.new()
    scopes = PartialScopeStack(scopes=(5,), variable=None)
    symbol = PartialScopedSymbol(symbol_id=1, scopes=scopes)
    lhs = PartialSymbolStack(symbols=(symbol,), variable=None)
    rhs = PartialSymbolStack(symbols=(symbol,), variable=None)

    result = lhs.unify(rhs, symbol_bindings, scope_bindings)

    assert result.symbols[0].symbol_id == 1
    assert result.symbols[0].scopes == scopes
    assert symbol_bindings.bindings == {}
    assert scope_bindings.bindings == {}


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_symbol_stack_unify_symbol_mismatch_raises() -> None:
    scope_bindings = PartialScopeStackBindings.new()
    symbol_bindings = PartialSymbolStackBindings.new()
    lhs = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=1, scopes=None),),
        variable=None,
    )
    rhs = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=2, scopes=None),),
        variable=None,
    )

    with pytest.raises(PartialPathResolutionError) as excinfo:
        lhs.unify(rhs, symbol_bindings, scope_bindings)

    exception = excinfo.value  # type: ignore[attr-defined]
    assert exception.code == PartialPathResolutionErrorCode.SYMBOL_STACK_UNSATISFIED


@pytest.mark.blends_test_group("stack_unittesting")
def test_concatenate_undoes_join_push_symbol() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    join = view.nid_to_index["2"]
    symbol_id = view.symbols.intern("x")

    left = PartialPath(
        start_node_index=0,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack(
            symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=None),),
            variable=None,
        ),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right = PartialPath(
        start_node_index=join,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )

    result = left.concatenate(view, right)

    assert result.symbol_stack_precondition == PartialSymbolStack.empty()
    assert result.symbol_stack_postcondition == PartialSymbolStack.empty()


@pytest.mark.blends_test_group("stack_unittesting")
def test_concatenate_undoes_join_pop_symbol() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    join = view.nid_to_index["2"]
    symbol_id = view.symbols.intern("x")

    left = PartialPath(
        start_node_index=0,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right = PartialPath(
        start_node_index=join,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack(
            symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=None),),
            variable=None,
        ),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )

    result = left.concatenate(view, right)

    assert result.symbol_stack_precondition == PartialSymbolStack.empty()
    assert result.symbol_stack_postcondition == PartialSymbolStack.empty()


@pytest.mark.blends_test_group("stack_unittesting")
def test_concatenate_incorrect_source_node_raises() -> None:
    graph = SyntaxGraph()
    view = _view_for_graph(graph)
    left = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right = PartialPath(
        start_node_index=1,
        end_node_index=1,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )

    with pytest.raises(PartialPathResolutionError) as excinfo:
        left.concatenate(view, right)

    exception = excinfo.value  # type: ignore[attr-defined]
    assert exception.code == PartialPathResolutionErrorCode.INCORRECT_SOURCE_NODE


@pytest.mark.blends_test_group("stack_unittesting")
def test_concatenate_unbound_symbol_stack_var_raises() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    join = view.nid_to_index["2"]
    symbol_id = view.symbols.intern("x")
    var = SymbolStackVariable.initial()

    left = PartialPath(
        start_node_index=0,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack(
            symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=None),),
            variable=None,
        ),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right = PartialPath(
        start_node_index=join,
        end_node_index=join,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.from_variable(var),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )

    with pytest.raises(PartialPathResolutionError) as excinfo:
        left.concatenate(view, right)

    exception = excinfo.value  # type: ignore[attr-defined]
    assert exception.code == PartialPathResolutionErrorCode.UNBOUND_SYMBOL_STACK_VARIABLE


@pytest.mark.blends_test_group("stack_unittesting")
def test_concatenate_unbound_scope_stack_var_raises() -> None:
    graph = SyntaxGraph()
    view = _view_for_graph(graph)
    var = ScopeStackVariable.initial()

    left = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.from_variable(var),
    )

    with pytest.raises(PartialPathResolutionError) as excinfo:
        left.concatenate(view, right)

    exception = excinfo.value  # type: ignore[attr-defined]
    assert exception.code == PartialPathResolutionErrorCode.UNBOUND_SCOPE_STACK_VARIABLE


@pytest.mark.blends_test_group("stack_unittesting")
def test_ensure_no_overlapping_variables_offsets_rhs() -> None:
    expected_var_id = 2
    var = SymbolStackVariable.initial()
    lhs = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.from_variable(var),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    rhs = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.from_variable(var),
        symbol_stack_postcondition=PartialSymbolStack.empty(),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )

    updated_rhs = rhs.ensure_no_overlapping_variables(lhs)

    assert updated_rhs.symbol_stack_precondition.variable is not None
    assert updated_rhs.symbol_stack_precondition.variable.as_int() == expected_var_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.for_each_statement import (
    build_for_each_statement_node,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    class_childs = adj_ast(graph, args.n_id)
    body_id = match_ast_d(graph, args.n_id, "block") or class_childs[0]
    var_node = match_ast_d(graph, args.n_id, "variable_declaration") or class_childs[-1]
    in_node = match_ast_d(graph, args.n_id, "in")

    if graph.nodes[var_node]["label_type"] == "variable_declaration" and (
        s_id := match_ast_d(graph, var_node, "identifier")
    ):
        var_node = s_id

    if in_node:
        iterable_item = class_childs[class_childs.index(in_node) + 1]
        return build_for_each_statement_node(args, var_node, iterable_item, body_id)

    return build_for_statement_node(args, var_node, None, None, body_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    var_id = n_attrs["label_field_left"]
    val_id = n_attrs.get("label_field_right")
    type_id = n_attrs.get("label_field_type")

    if graph.nodes[var_id]["label_type"] == "identifier" or not val_id:
        var_name = node_to_str(graph, n_attrs["label_field_left"])
        var_type = None
        if type_id:
            var_type = node_to_str(graph, type_id)
        return build_variable_declaration_node(args, var_name, var_type, val_id)

    return build_assignment_node(args, var_id, val_id, type_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    match_childs = match_ast(graph, args.n_id, "block", "binary_expression")

    block_id = match_childs.get("block")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[-1]

    if block_id and graph.nodes[block_id]["label_type"] == "expression_statement":
        block_id = adj_ast(graph, block_id)[0]

    conditional_node = match_childs.get("binary_expression")

    return build_while_statement_node(args, block_id, conditional_node)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_new_expression_node(
    args: SyntaxGraphArgs,
    constructor_id: NId,
    arguments_id: NId | None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        constructor_id=constructor_id,
        label_type="NewExpression",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(constructor_id)),
        label_ast="AST",
    )

    if arguments_id and args.ast_graph.nodes[arguments_id].get("label_type") != "const":
        args.syntax_graph.update_node_attribute(args.n_id, "arguments_id", arguments_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(arguments_id)),
            label_ast="AST",
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition_id = (
        n_attrs.get("label_field_condition")
        or match_ast_d(graph, args.n_id, "simple_identifier")
        or match_ast_d(graph, args.n_id, "call_expression")
        or adj_ast(graph, args.n_id)[0]
    )
    initializer_id = n_attrs.get("label_field_bound_identifier")
    if (
        condition_id is not None
        and graph.nodes[condition_id]["label_type"] == "value_binding_pattern"
        and (
            (try_id := match_ast_d(graph, args.n_id, "try_expression"))
            or (call_expr := match_ast_d(graph, args.n_id, "call_expression"))
        )
    ):
        condition_id = try_id or call_expr

    statements = match_ast_group_d(graph, args.n_id, "statements")

    true_id = None
    false_id = None

    expected_statements_for_if_with_else = 2
    if len(statements) == expected_statements_for_if_with_else:
        true_id = statements[0]
        false_id = statements[1]
    elif len(statements) == 1:
        true_id = statements[0]

    if false_statement := match_ast_d(graph, args.n_id, "if_statement"):
        false_id = false_statement

    return build_if_node(
        args, condition_id, true_id, false_id, initializer_id if initializer_id else None
    )



from pathlib import Path
from unittest.mock import patch

from blends import get_graphs_from_path
from blends.models import SyntaxGraph
from blends.stack.forward_stitcher import ForwardStitcherConfig
from blends.stack.partial_path import (
    PartialPath,
    PartialPathDatabase,
    PartialPathFileRecord,
    PartialPathIndexRequest,
    PartialPathLimits,
    PartialScopeStack,
    PartialSymbolStack,
    PerFilePartialPathStats,
    StackGraphViewMetadata,
    index_partial_paths_for_file,
)
from blends.stack.partial_path.variables import ScopeStackVariable, SymbolStackVariable
from blends.stack.single_file_resolver import (
    SingleFileResolutionRequest,
    resolve_definitions_single_file,
)
from blends.stack.view import StackGraphView

INDEX_MAX_WORK = 50_000
QUERY_MAX_PHASES = 100
QUERY_MAX_WORK = 10_000
STACK_GRAPH_TEST_DATA_DIR = Path(__file__).parent.parent / "data" / "test_files" / "stack_graph"


def setup_multi_file_database(
    files: dict[str, int],
) -> tuple[dict[int, StackGraphView], PartialPathDatabase]:
    database = PartialPathDatabase()
    views: dict[int, StackGraphView] = {}
    for filename, file_handle in files.items():
        path = STACK_GRAPH_TEST_DATA_DIR / filename
        logical = f"test/data/test_files/stack_graph/{filename}"
        with patch(
            "blends.ctx.ctx.has_feature_flag",
            side_effect=lambda flag: flag == "StackGraph",
        ):
            graphs = get_graphs_from_path(path)
        assert graphs.syntax_graph is not None
        views[file_handle] = StackGraphView.from_syntax_graph(graphs.syntax_graph, path=logical)
        index_partial_paths_for_file(
            graphs.syntax_graph,
            request=PartialPathIndexRequest(
                file_path=logical,
                file_handle=file_handle,
                limits=PartialPathLimits(max_work_per_phase=INDEX_MAX_WORK),
            ),
            database=database,
        )
    return views, database


def view_for_graph(graph: SyntaxGraph, *, path: str = "test.py") -> StackGraphView:
    return StackGraphView.from_syntax_graph(graph, path=path)


def make_path(start: int, end: int) -> PartialPath:
    symbol_var = SymbolStackVariable.initial()
    scope_var = ScopeStackVariable.initial()
    return PartialPath(
        start_node_index=start,
        end_node_index=end,
        symbol_stack_precondition=PartialSymbolStack.from_variable(symbol_var),
        symbol_stack_postcondition=PartialSymbolStack.from_variable(symbol_var),
        scope_stack_precondition=PartialScopeStack.from_variable(scope_var),
        scope_stack_postcondition=PartialScopeStack.from_variable(scope_var),
    )


def record(
    paths: tuple[PartialPath, ...],
    *,
    file_handle: int = 1,
    file_path: str = "test.py",
    symbol_id_to_string: tuple[str, ...] = ("x",),
) -> PartialPathFileRecord:
    metadata = StackGraphViewMetadata(
        file_handles=(file_path,),
        symbol_id_to_string=symbol_id_to_string,
    )
    return PartialPathFileRecord(
        file_handle=file_handle,
        file_path=file_path,
        view_snapshot=metadata,
        paths=paths,
        stats=PerFilePartialPathStats(),
    )


def setup_stack_resolution(test_file: str) -> tuple[StackGraphView, PartialPathFileRecord]:
    test_file_path = STACK_GRAPH_TEST_DATA_DIR / test_file
    file_path = f"test/data/test_files/stack_graph/{test_file}"

    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        graphs = get_graphs_from_path(test_file_path)

    assert graphs.syntax_graph is not None

    view = StackGraphView.from_syntax_graph(graphs.syntax_graph, path=file_path)

    database = PartialPathDatabase()
    per_file = index_partial_paths_for_file(
        graphs.syntax_graph,
        request=PartialPathIndexRequest(
            file_path=file_path,
            file_handle=1,
            limits=PartialPathLimits(max_work_per_phase=INDEX_MAX_WORK),
        ),
        database=database,
    )
    assert per_file.paths, "Expected at least one partial path"

    file_record = database.get_file(1)
    assert file_record is not None

    return view, file_record


def find_symbol_nodes(
    view: StackGraphView,
    *,
    symbol_id: int,
    is_definition: bool,
) -> list[int]:
    matches: list[int] = []
    flags = view.node_is_definition if is_definition else view.node_is_reference
    for node_index, flag in enumerate(flags):
        if not flag:
            continue
        if view.node_symbol_id[node_index] == symbol_id:
            matches.append(node_index)
    return matches


def resolve_symbol(
    view: StackGraphView,
    file_record: PartialPathFileRecord,
    symbol: str,
    ref_index: int = 0,
) -> list[int]:
    symbol_id = view.symbols.symbol_to_id.get(symbol)
    if symbol_id is None:
        return []

    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    if ref_index >= len(ref_nodes):
        return []

    request = SingleFileResolutionRequest(
        file_handle=1,
        record=file_record,
        ref_node_index=ref_nodes[ref_index],
        config=ForwardStitcherConfig(
            max_phases=QUERY_MAX_PHASES,
            max_work_per_phase=QUERY_MAX_WORK,
            max_total_work=QUERY_MAX_WORK,
        ),
    )
    result = resolve_definitions_single_file(view, request=request)
    assert isinstance(result, list)
    return result


def count_defs_refs(view: StackGraphView, symbol: str) -> tuple[int, int]:
    symbol_id = view.symbols.symbol_to_id.get(symbol)
    if symbol_id is None:
        return (0, 0)
    defs = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
    refs = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    return (len(defs), len(refs))


def get_definition_precedences(view: StackGraphView, symbol: str) -> list[int]:
    symbol_id = view.symbols.symbol_to_id.get(symbol)
    if symbol_id is None:
        return []
    def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
    return [view.node_precedence[node_index] for node_index in def_nodes]



from blends.models import (
    NId,
)
from blends.syntax.builders.for_each_statement import (
    build_for_each_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    node = graph.nodes[args.n_id]
    var_node = node["label_field_left"]
    iterable_item = node["label_field_right"]

    return build_for_each_statement_node(args, var_node, iterable_item, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    name: str = ""
    node_ids: set[NId] = set[NId]()

    for idx, node_id in match_ast(graph, args.n_id).items():
        if not node_id:
            continue

        if idx == "__0__":
            name = node_to_str(graph, node_id)
            continue

        if graph.nodes[node_id]["label_type"] == "parenthesized_expression":
            node_ids.add(adj_ast(graph, node_id)[1])
        else:
            node_ids.add(node_id)

    return build_import_global_node(args, name, node_ids, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    expr_id = adj_ast(args.ast_graph, args.n_id)[0]
    return args.generic(args.fork_n_id(expr_id))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.cfg.dispatchers.multifile import (
    get_deterministic_path_id,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)


def build(args: SyntaxCfgArgs) -> NId:
    if args.is_multifile and (f_id := get_deterministic_path_id(args)):
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(f_id, args.nxt_id)),
            label_cfg="CFG",
        )
    else:
        for c_id in adj_ast(args.graph, args.n_id):
            args.graph.add_edge(
                args.n_id,
                args.generic(args.fork(c_id, args.nxt_id)),
                label_cfg="CFG",
            )
    return args.n_id



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.builders.for_each_statement import (
    build_for_each_statement_node,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


def _make_args(
    *, syntax_graph: SyntaxGraph, ast_graph: Graph, n_id: NId, metadata: SyntaxMetadata
) -> SyntaxGraphArgs:
    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id=n_id,
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_assignment_emits_pop_symbol_and_scope_edge_for_identifier_target() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"]}
    syntax_graph.add_node("1", label_type="File")

    ast_graph.add_node("10", label_type="identifier", label_text="x")
    ast_graph.add_node("11", label_type="integer", label_text="1")

    args = _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="2", metadata=metadata)
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_assignment_node(args, var_id="10", val_id="11", operator=None)

    assert syntax_graph.nodes["2"][STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    assert syntax_graph.nodes["2"][STACK_GRAPH_SYMBOL] == "x"
    assert syntax_graph.has_edge("1", "2")


@pytest.mark.blends_test_group("stack_unittesting")
def test_for_each_emits_pop_symbol_and_scope_edge_for_identifier_target() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"]}
    syntax_graph.add_node("1", label_type="File")

    ast_graph.add_node("10", label_type="identifier", label_text="x")
    ast_graph.add_node("11", label_type="identifier", label_text="xs")

    args = _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="2", metadata=metadata)
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_for_each_statement_node(args, var_node="10", iterable_item="11", block_id=None)

    assert syntax_graph.nodes["2"][STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    assert syntax_graph.nodes["2"][STACK_GRAPH_SYMBOL] == "x"
    assert syntax_graph.has_edge("1", "2")


@pytest.mark.blends_test_group("stack_unittesting")
def test_for_statement_emits_pop_symbol_and_scope_edge_for_identifier_initializer() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"]}
    syntax_graph.add_node("1", label_type="File")

    ast_graph.add_node("10", label_type="identifier", label_text="x")
    ast_graph.add_node("11", label_type="identifier", label_text="xs")
    ast_graph.add_node("12", label_type="block", label_text="")

    args = _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="2", metadata=metadata)
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_for_statement_node(
            args, initializer_node="10", condition_node="11", update_node=None, body_node="12"
        )

    assert syntax_graph.nodes["2"][STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    assert syntax_graph.nodes["2"][STACK_GRAPH_SYMBOL] == "x"
    assert syntax_graph.has_edge("1", "2")


@pytest.mark.blends_test_group("stack_unittesting")
def test_multiple_assignments_get_incrementing_precedence() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"]}
    syntax_graph.add_node("1", label_type="File")

    ast_graph.add_node("10", label_type="identifier", label_text="x")
    ast_graph.add_node("11", label_type="integer", label_text="1")
    ast_graph.add_node("20", label_type="identifier", label_text="x")
    ast_graph.add_node("21", label_type="integer", label_text="2")

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        args1 = _make_args(
            syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="2", metadata=metadata
        )
        build_assignment_node(args1, var_id="10", val_id="11", operator=None)

        args2 = _make_args(
            syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="3", metadata=metadata
        )
        build_assignment_node(args2, var_id="20", val_id="21", operator=None)

    assert syntax_graph.has_edge("1", "2")
    assert syntax_graph.has_edge("1", "3")

    edge_data_first = syntax_graph["1"]["2"]
    edge_data_second = syntax_graph["1"]["3"]

    assert edge_data_first["precedence"] == 1
    assert edge_data_second["precedence"] == 1
    assert edge_data_first["precedence"] == edge_data_second["precedence"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_binding_sites_do_not_emit_stack_graph_when_feature_preview_disabled() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"]}
    syntax_graph.add_node("1", label_type="File")

    ast_graph.add_node("10", label_type="identifier", label_text="x")
    ast_graph.add_node("11", label_type="identifier", label_text="xs")
    ast_graph.add_node("12", label_type="block", label_text="")

    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_assignment_node(
            _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="2", metadata=metadata),
            var_id="10",
            val_id="11",
            operator=None,
        )
        build_for_each_statement_node(
            _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="3", metadata=metadata),
            var_node="10",
            iterable_item="11",
            block_id=None,
        )
        build_for_statement_node(
            _make_args(syntax_graph=syntax_graph, ast_graph=ast_graph, n_id="4", metadata=metadata),
            initializer_node="10",
            condition_node="11",
            update_node=None,
            body_node="12",
        )

    assert STACK_GRAPH_KIND not in syntax_graph.nodes["2"]
    assert STACK_GRAPH_KIND not in syntax_graph.nodes["3"]
    assert STACK_GRAPH_KIND not in syntax_graph.nodes["4"]
    assert not syntax_graph.has_edge("1", "2")
    assert not syntax_graph.has_edge("1", "3")
    assert not syntax_graph.has_edge("1", "4")



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.stack.node_helpers import (
    scope_node_attributes,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.scope_types import ScopeType
from blends.stack.view import (
    StackGraphView,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.builders.method_declaration import (
    build_method_declaration_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


def _make_args(
    *,
    syntax_graph: SyntaxGraph,
    n_id: NId,
    metadata: SyntaxMetadata,
    language: Language = Language.PYTHON,
) -> SyntaxGraphArgs:
    ast_graph = Graph()

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=language,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id=n_id,
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_extracts_reference_to_definition_edges() -> None:
    syntax_graph = SyntaxGraph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"], "next_synthetic_id": 100}
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node("1", scope_node_attributes(is_exported=True))

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        def_args = _make_args(syntax_graph=syntax_graph, n_id="2", metadata=metadata)
        build_variable_declaration_node(
            def_args,
            variable_name="x",
            variable_type=None,
            value_id=None,
        )

        ref_args = _make_args(syntax_graph=syntax_graph, n_id="3", metadata=metadata)
        build_symbol_lookup_node(ref_args, symbol="x")

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    file_index = view.nid_to_index["1"]
    def_index = view.nid_to_index["2"]
    ref_index = view.nid_to_index["3"]

    assert view.node_kind[file_index] == StackGraphNodeKind.SCOPE.value
    assert view.node_kind[def_index] == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    assert view.node_kind[ref_index] == StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value

    assert (file_index, 0) in view.outgoing[ref_index]
    assert (def_index, 1) in view.outgoing[file_index]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_extracts_nested_scope_edges() -> None:
    syntax_graph = SyntaxGraph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"], "next_synthetic_id": 100}
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node("1", scope_node_attributes(is_exported=True))

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        class_args = _make_args(syntax_graph=syntax_graph, n_id="10", metadata=metadata)
        build_class_node(
            class_args,
            name="MyClass",
            block_id=None,
            attrl_ids=None,
            inherited_class=None,
            modifiers_id=None,
        )

        method_metadata: SyntaxMetadata = {
            "class_path": [],
            "scope_stack": ["1", "10"],
            "next_synthetic_id": 101,
        }
        method_args = _make_args(syntax_graph=syntax_graph, n_id="20", metadata=method_metadata)
        build_method_declaration_node(
            method_args,
            name="foo",
            direct_children={},
            list_children={},
        )

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    file_index = view.nid_to_index["1"]
    class_index = view.nid_to_index["10"]
    method_index = view.nid_to_index["20"]

    assert view.node_kind[file_index] == StackGraphNodeKind.SCOPE.value
    assert view.node_kind[class_index] == StackGraphNodeKind.SCOPE.value
    assert view.node_kind[method_index] == StackGraphNodeKind.SCOPE.value

    assert (file_index, 0) in view.outgoing[class_index]
    assert (class_index, 0) in view.outgoing[method_index]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_marks_exported_scopes() -> None:
    syntax_graph = SyntaxGraph()
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node("1", scope_node_attributes(is_exported=True))
    syntax_graph.add_node("2", label_type="Class")
    syntax_graph.update_node("2", scope_node_attributes(is_exported=False))

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    file_index = view.nid_to_index["1"]
    class_index = view.nid_to_index["2"]

    assert file_index in view.exported_scopes
    assert class_index not in view.exported_scopes


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_preserves_edge_precedence_values() -> None:
    syntax_graph = SyntaxGraph()
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node("1", scope_node_attributes(is_exported=True))

    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"], "next_synthetic_id": 100}
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        def_args_a = _make_args(syntax_graph=syntax_graph, n_id="2", metadata=metadata)
        build_variable_declaration_node(
            def_args_a,
            variable_name="a",
            variable_type=None,
            value_id=None,
        )
        def_args_b = _make_args(syntax_graph=syntax_graph, n_id="3", metadata=metadata)
        build_variable_declaration_node(
            def_args_b,
            variable_name="b",
            variable_type=None,
            value_id=None,
        )

    syntax_graph.add_edge("1", "2", precedence=10)
    syntax_graph.add_edge("1", "3", precedence=1)

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    file_index = view.nid_to_index["1"]
    def_index_a = view.nid_to_index["2"]
    def_index_b = view.nid_to_index["3"]

    outgoing = set(view.outgoing[file_index])
    assert (def_index_a, 10) in outgoing
    assert (def_index_b, 1) in outgoing


@pytest.mark.blends_test_group("stack_unittesting")
def test_python_file_scope_exported_with_scope_type() -> None:
    syntax_graph = SyntaxGraph()
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": [], "next_synthetic_id": 100}

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        file_args = _make_args(syntax_graph=syntax_graph, n_id="1", metadata=metadata)
        build_file_node(file_args, c_ids=[])

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    file_index = view.nid_to_index["1"]

    assert file_index in view.exported_scopes
    assert view.node_scope_type[file_index] == ScopeType.FILE.value


@pytest.mark.blends_test_group("stack_unittesting")
def test_python_function_scope_not_exported_with_scope_type() -> None:
    syntax_graph = SyntaxGraph()
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node(
        "1", scope_node_attributes(is_exported=True, scope_type=ScopeType.FILE)
    )
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"], "next_synthetic_id": 100}

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        method_args = _make_args(syntax_graph=syntax_graph, n_id="10", metadata=metadata)
        build_method_declaration_node(
            method_args, name="my_func", direct_children={}, list_children={}
        )

    view = StackGraphView.from_syntax_graph(syntax_graph, path="test.py")
    method_index = view.nid_to_index["10"]

    assert method_index not in view.exported_scopes
    assert view.node_scope_type[method_index] == ScopeType.FUNCTION.value


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("access_modifiers", "expected_exported"),
    [
        ("public", True),
        (None, False),
    ],
)
def test_java_class_scope_exported_based_on_visibility(  # type: ignore[misc]
    access_modifiers: str | None,
    *,
    expected_exported: bool,
) -> None:
    syntax_graph = SyntaxGraph()
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.update_node(
        "1", scope_node_attributes(is_exported=True, scope_type=ScopeType.FILE)
    )
    metadata: SyntaxMetadata = {"class_path": [], "scope_stack": ["1"], "next_synthetic_id": 100}

    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        class_args = _make_args(
            syntax_graph=syntax_graph,
            n_id="10",
            metadata=metadata,
            language=Language.JAVA,
        )
        build_class_node(
            class_args,
            name="MyClass",
            block_id=None,
            attrl_ids=None,
            access_modifiers=access_modifiers,
        )

    view = StackGraphView.from_syntax_graph(syntax_graph, path="Test.java")
    class_index = view.nid_to_index["10"]

    assert (class_index in view.exported_scopes) is expected_exported
    assert view.node_scope_type[class_index] == ScopeType.CLASS.value



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expression_id = graph.nodes[args.n_id]["label_field_condition"]
    expression = node_to_str(graph, expression_id)
    member_id = (
        match_ast_d(graph, args.n_id, "member_binding_expression")
        or match_ast_d(graph, args.n_id, "element_binding_expression")
        or adj_ast(graph, args.n_id)[-1]
    )
    member = node_to_str(graph, member_id).strip(".")
    return build_member_access_node(args, member, expression, expression_id)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_switch_statement_node(
    args: SyntaxGraphArgs,
    body_id: NId,
    value_id: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        block_id=body_id,
        value_id=value_id,
        label_type="SwitchStatement",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(body_id)),
        label_ast="AST",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(value_id)),
        label_ast="AST",
    )

    return args.n_id



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from blends.stack.partial_path.minimal_paths import (
    PartialPathLimits,
    compute_minimal_partial_paths_in_file,
)
from blends.stack.view import StackGraphView

if TYPE_CHECKING:
    from blends.models import SyntaxGraph
    from blends.stack.partial_path.minimal_paths import PerFilePartialPaths
    from blends.stack.partial_path_database import FileHandle, PartialPathDatabase


@dataclass(frozen=True, slots=True)
class PartialPathIndexRequest:
    file_path: str
    file_handle: FileHandle
    limits: PartialPathLimits | None = None
    record_edges: bool = True


def index_partial_paths_for_file(
    syntax_graph: SyntaxGraph,
    request: PartialPathIndexRequest,
    database: PartialPathDatabase,
) -> PerFilePartialPaths:
    view = StackGraphView.from_syntax_graph(syntax_graph, path=request.file_path)
    result = compute_minimal_partial_paths_in_file(
        view,
        request.file_path,
        limits=request.limits,
        record_edges=request.record_edges,
    )
    database.put_file(
        file_handle=request.file_handle,
        file_path=request.file_path,
        view=view,
        partial_paths=result,
    )
    return result



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    val_id = n_attrs["label_field_value"]

    childs = match_ast(args.ast_graph, args.n_id, "=")
    var_name = node_to_str(args.ast_graph, var_id) if (var_id := childs.get("__0__")) else "unnamed"

    return build_variable_declaration_node(args, var_name, None, val_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    arrow_id = args.ast_graph.nodes[args.n_id]
    block_id = arrow_id["label_field_body"]
    params = arrow_id.get("label_field_parameter") or arrow_id.get("label_field_parameters")

    direct: DirectChildren = {"block_id": block_id}
    if params:
        direct["parameters_id"] = params

    return build_method_declaration_node(args, None, direct, {})



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = match_ast(graph, args.n_id)
    operator = ""
    if operator_n_id := c_ids.get("__0__"):
        operator = graph.nodes[operator_n_id].get("label_text", "")
    operand = c_ids.get("__1__") or args.n_id

    return build_unary_expression_node(args, operator, operand)



import pytest

from blends.models import (
    Graph,
    NId,
)
from blends.syntax.readers.python.parameter import (
    _extract_annotation_class_names,
)


def _make_identifier_graph(name: str) -> tuple[Graph, NId]:
    g = Graph()
    g.add_node("1", label_type="identifier", label_text=name)
    return g, "1"


def _make_attribute_graph(obj: str, attr: str) -> tuple[Graph, NId]:
    g = Graph()
    g.add_node("1", label_type="attribute", label_field_attribute="3")
    g.add_node("2", label_type="identifier", label_text=obj)
    g.add_node("3", label_type="identifier", label_text=attr)
    g.add_edge("1", "2", label_ast="AST")
    g.add_edge("1", "3", label_ast="AST")
    return g, "1"


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_simple_class_name() -> None:
    g, root = _make_identifier_graph("Foo")
    assert _extract_annotation_class_names(g, root) == {"Foo"}

    g2, root2 = _make_identifier_graph("int")
    assert _extract_annotation_class_names(g2, root2) == {"int"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_none_returns_empty() -> None:
    g, root = _make_identifier_graph("None")
    assert _extract_annotation_class_names(g, root) == set()


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_forward_ref_string() -> None:
    g = Graph()
    g.add_node("1", label_type="string", label_text='"Foo"')
    assert _extract_annotation_class_names(g, "1") == {"Foo"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_qualified_forward_ref_string_returns_last_segment() -> None:
    g = Graph()
    g.add_node("1", label_type="string", label_text='"module.Foo"')
    assert _extract_annotation_class_names(g, "1") == {"Foo"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_qualified_name_returns_last_segment() -> None:
    g, root = _make_attribute_graph("pkg", "Foo")
    assert _extract_annotation_class_names(g, root) == {"Foo"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_union_with_none() -> None:
    g = Graph()
    g.add_node("1", label_type="binary_operator")
    g.add_node("2", label_type="identifier", label_text="Counter")
    g.add_node("3", label_type="|", label_text="|")
    g.add_node("4", label_type="identifier", label_text="None")
    g.add_edge("1", "2", label_ast="AST")
    g.add_edge("1", "3", label_ast="AST")
    g.add_edge("1", "4", label_ast="AST")
    result = _extract_annotation_class_names(g, "1")
    assert result == {"Counter"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_type_wrapper_is_transparent() -> None:
    g = Graph()
    g.add_node("1", label_type="type")
    g.add_node("2", label_type="identifier", label_text="Counter")
    g.add_edge("1", "2", label_ast="AST")
    assert _extract_annotation_class_names(g, "1") == {"Counter"}


@pytest.mark.blends_test_group("stack_unittesting")
def test_extract_unknown_node_type_returns_empty() -> None:
    g = Graph()
    g.add_node("1", label_type="tuple_type", label_text="tuple[int, str]")
    assert _extract_annotation_class_names(g, "1") == set()



from dataclasses import dataclass

from blends.models import (
    NId,
    SyntaxGraph,
)


@dataclass(frozen=True, slots=True)
class Edge:
    source: NId
    sink: NId
    precedence: int = 0


def add_edge(graph: SyntaxGraph, graph_edge: Edge) -> None:
    source = graph_edge.source
    sink = graph_edge.sink
    precedence = graph_edge.precedence if isinstance(graph_edge.precedence, int) else 0
    if graph.has_edge(source, sink):
        existing_data = graph.get_edge_data(source, sink, default={})
        existing_precedence = existing_data.get("precedence", 0)
        if isinstance(existing_precedence, int):
            precedence = max(existing_precedence, precedence)
    graph.add_edge(source, sink, precedence=precedence)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.javascript import (
    arguments as javascript_arguments,
)
from blends.syntax.readers.javascript import (
    array_node as javascript_array,
)
from blends.syntax.readers.javascript import (
    arrow_function as javascript_arrow_function,
)
from blends.syntax.readers.javascript import (
    assignment_expression as javascript_assignment_expression,
)
from blends.syntax.readers.javascript import (
    await_expression as javascript_await_expression,
)
from blends.syntax.readers.javascript import (
    binary_expression as javascript_binary_expression,
)
from blends.syntax.readers.javascript import (
    boolean_literal as javascript_boolean_literal,
)
from blends.syntax.readers.javascript import (
    break_statement as javascript_break_statement,
)
from blends.syntax.readers.javascript import (
    call_expression as javascript_call_expression,
)
from blends.syntax.readers.javascript import (
    catch_clause as javascript_catch_clause,
)
from blends.syntax.readers.javascript import (
    class_body as javascript_class_body,
)
from blends.syntax.readers.javascript import (
    class_declaration as javascript_class_declaration,
)
from blends.syntax.readers.javascript import (
    comment as javascript_comment,
)
from blends.syntax.readers.javascript import (
    debugger_statement as javascript_debugger_statement,
)
from blends.syntax.readers.javascript import (
    do_statement as javascript_do_statement,
)
from blends.syntax.readers.javascript import (
    else_clause as javascript_else_clause,
)
from blends.syntax.readers.javascript import (
    execution_block as javascript_execution_block,
)
from blends.syntax.readers.javascript import (
    export_statement as javascript_export_statement,
)
from blends.syntax.readers.javascript import (
    expression_statement as javascript_expression_statement,
)
from blends.syntax.readers.javascript import (
    finally_clause as javascript_finally_clause,
)
from blends.syntax.readers.javascript import (
    for_each_statement as javascript_for_each_statement,
)
from blends.syntax.readers.javascript import (
    for_statement as javascript_for_statement,
)
from blends.syntax.readers.javascript import (
    identifier as javascript_identifier,
)
from blends.syntax.readers.javascript import (
    if_statement as javascript_if_statement,
)
from blends.syntax.readers.javascript import (
    import_global as javascript_import_global,
)
from blends.syntax.readers.javascript import (
    import_module as javascript_import_module,
)
from blends.syntax.readers.javascript import (
    import_node as javascript_import,
)
from blends.syntax.readers.javascript import (
    jsx_attribute as javascript_jsx_attribute,
)
from blends.syntax.readers.javascript import (
    jsx_element as javascript_jsx_element,
)
from blends.syntax.readers.javascript import (
    jsx_self_closing_element as javascript_jsx_self_closing_element,
)
from blends.syntax.readers.javascript import (
    member_expression as javascript_member_expression,
)
from blends.syntax.readers.javascript import (
    method_declaration as javascript_method_declaration,
)
from blends.syntax.readers.javascript import (
    new_expression as javascript_new_expression,
)
from blends.syntax.readers.javascript import (
    null_literal as javascript_null_literal,
)
from blends.syntax.readers.javascript import (
    number_literal as javascript_number_literal,
)
from blends.syntax.readers.javascript import (
    object as javascript_object,
)
from blends.syntax.readers.javascript import (
    pair as javascript_pair,
)
from blends.syntax.readers.javascript import (
    parameter_list as javascript_parameter_list,
)
from blends.syntax.readers.javascript import (
    parenthesized_expression as javascript_parenthesized_expression,
)
from blends.syntax.readers.javascript import (
    program as javascript_program,
)
from blends.syntax.readers.javascript import (
    rest_pattern as javascript_rest_pattern,
)
from blends.syntax.readers.javascript import (
    return_statement as javascript_return_statement,
)
from blends.syntax.readers.javascript import (
    spread_element as javascript_spreadt_element,
)
from blends.syntax.readers.javascript import (
    string_literal as javascript_string_literal,
)
from blends.syntax.readers.javascript import (
    subscript_expression as javascript_subscript_expression,
)
from blends.syntax.readers.javascript import (
    switch_body as javascript_switch_body,
)
from blends.syntax.readers.javascript import (
    switch_section as javascript_switch_section,
)
from blends.syntax.readers.javascript import (
    switch_statement as javascript_switch_statement,
)
from blends.syntax.readers.javascript import (
    template_string as javascript_template_string,
)
from blends.syntax.readers.javascript import (
    ternary_expression as javascript_ternary_expression,
)
from blends.syntax.readers.javascript import (
    this_node as javascript_this,
)
from blends.syntax.readers.javascript import (
    throw_statement as javascript_throw_statement,
)
from blends.syntax.readers.javascript import (
    try_statement as javascript_try_statement,
)
from blends.syntax.readers.javascript import (
    unary_expression as javascript_unary_expression,
)
from blends.syntax.readers.javascript import (
    update_expression as javascript_update_expression,
)
from blends.syntax.readers.javascript import (
    variable_declaration as javascript_variable_declaration,
)
from blends.syntax.readers.javascript import (
    variable_declarator as javascript_variable_declarator,
)
from blends.syntax.readers.javascript import (
    while_statement as javascript_while_statement,
)
from blends.syntax.readers.javascript import (
    yield_expression as javascript_yield_expression,
)

JAVASCRIPT_DISPATCHERS: Dispatcher = {
    "arguments": javascript_arguments.reader,
    "array": javascript_array.reader,
    "arrow_function": javascript_arrow_function.reader,
    "assignment_expression": javascript_assignment_expression.reader,
    "augmented_assignment_expression": javascript_assignment_expression.reader,
    "await_expression": javascript_await_expression.reader,
    "jsx_element": javascript_jsx_element.reader,
    "jsx_fragment": javascript_jsx_element.reader,
    "jsx_self_closing_element": javascript_jsx_self_closing_element.reader,
    "jsx_opening_element": javascript_jsx_element.reader,
    "jsx_attribute": javascript_jsx_attribute.reader,
    "binary_expression": javascript_binary_expression.reader,
    "false": javascript_boolean_literal.reader,
    "true": javascript_boolean_literal.reader,
    "break_statement": javascript_break_statement.reader,
    "call_expression": javascript_call_expression.reader,
    "catch_clause": javascript_catch_clause.reader,
    "class_body": javascript_class_body.reader,
    "class_declaration": javascript_class_declaration.reader,
    "comment": javascript_comment.reader,
    "debugger_statement": javascript_debugger_statement.reader,
    "do_statement": javascript_do_statement.reader,
    "else_clause": javascript_else_clause.reader,
    "export_statement": javascript_export_statement.reader,
    "expression_statement": javascript_expression_statement.reader,
    "identifier": javascript_identifier.reader,
    "property_identifier": javascript_identifier.reader,
    "shorthand_property_identifier": javascript_identifier.reader,
    "shorthand_property_identifier_pattern": javascript_identifier.reader,
    "if_statement": javascript_if_statement.reader,
    "import": javascript_import.reader,
    "import_statement": javascript_import_global.reader,
    "import_specifier": javascript_import_module.reader,
    "finally_clause": javascript_finally_clause.reader,
    "for_in_statement": javascript_for_each_statement.reader,
    "for_statement": javascript_for_statement.reader,
    "member_expression": javascript_member_expression.reader,
    "function": javascript_method_declaration.reader,
    "function_declaration": javascript_method_declaration.reader,
    "generator_function_declaration": javascript_method_declaration.reader,
    "method_definition": javascript_method_declaration.reader,
    "function_expression": javascript_method_declaration.reader,
    "new_expression": javascript_new_expression.reader,
    "number": javascript_number_literal.reader,
    "null": javascript_null_literal.reader,
    "object": javascript_object.reader,
    "object_pattern": javascript_object.reader,
    "pair": javascript_pair.reader,
    "formal_parameters": javascript_parameter_list.reader,
    "parenthesized_expression": javascript_parenthesized_expression.reader,
    "rest_pattern": javascript_rest_pattern.reader,
    "spread_element": javascript_spreadt_element.reader,
    "program": javascript_program.reader,
    "return_statement": javascript_return_statement.reader,
    "statement_block": javascript_execution_block.reader,
    "regex": javascript_string_literal.reader,
    "string": javascript_string_literal.reader,
    "undefined": javascript_string_literal.reader,
    "subscript_expression": javascript_subscript_expression.reader,
    "switch_body": javascript_switch_body.reader,
    "switch_case": javascript_switch_section.reader,
    "switch_default": javascript_switch_section.reader,
    "switch_statement": javascript_switch_statement.reader,
    "template_string": javascript_template_string.reader,
    "ternary_expression": javascript_ternary_expression.reader,
    "this": javascript_this.reader,
    "throw_statement": javascript_throw_statement.reader,
    "try_statement": javascript_try_statement.reader,
    "unary_expression": javascript_unary_expression.reader,
    "update_expression": javascript_update_expression.reader,
    "variable_declaration": javascript_variable_declaration.reader,
    "variable_declarator": javascript_variable_declarator.reader,
    "lexical_declaration": javascript_variable_declaration.reader,
    "while_statement": javascript_while_statement.reader,
    "yield_expression": javascript_yield_expression.reader,
}



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_named_argument_node(
    args: SyntaxGraphArgs,
    arg_name: str | None,
    val_id: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        value_id=val_id,
        label_type="NamedArgument",
    )

    if arg_name:
        args.syntax_graph.update_node_attribute(args.n_id, "argument_name", arg_name)

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(val_id)),
        label_ast="AST",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    element_id = match_ast_d(graph, args.n_id, "const_element") or args.n_id
    c_ids = match_ast(graph, element_id)

    var_id = c_ids.get("__0__")
    var_name = (nodes[var_id].get("label_text") if var_id else None) or node_to_str(
        graph, args.n_id
    )
    val_id = c_ids.get("__2__")

    return build_variable_declaration_node(args, var_name, None, val_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.execution_block import (
    build_execution_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    _, *c_ids, _ = adj_ast(graph, args.n_id)  # do not consider { }

    return build_execution_block_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] != ";"),
    )



from __future__ import annotations

SENTINEL: int = -1
MAX_RETRIES: int = 3
DEFAULT_TIMEOUT: float = 30.0
MAX_POOL_SIZE: int = 64
CHUNK_SIZE: int = 256
BASE_SCORE: int = 100
BACKOFF_MS: int = 100
HTTP_OK: int = 200
HTTP_NOT_FOUND: int = 404
HTTP_SERVER_ERROR: int = 500
HTTP_UNAVAILABLE: int = 503


class CoreError:
    code: int = 0
    message: str = ""

    def is_fatal(self: "CoreError") -> bool:
        return self.code >= HTTP_SERVER_ERROR

    def is_client_error(self: "CoreError") -> bool:
        return 400 <= self.code < HTTP_SERVER_ERROR

    def describe(self: "CoreError") -> str:
        return "code=" + str(self.code) + " " + self.message

    def is_ok(self: "CoreError") -> bool:
        return self.code == 0


class RetryPolicy:
    max_attempts: int = MAX_RETRIES
    backoff_ms: int = BACKOFF_MS
    jitter: bool = False

    def should_retry(self: "RetryPolicy", attempt: int) -> bool:
        return attempt < self.max_attempts

    def delay_ms(self: "RetryPolicy", attempt: int) -> int:
        factor: int = attempt + 1
        return self.backoff_ms * factor

    def total_budget_ms(self: "RetryPolicy") -> int:
        total: int = 0
        idx: int = 0
        while idx < self.max_attempts:
            total = total + self.delay_ms(idx)
            idx = idx + 1
        return total



class ConnectionConfig:
    host: str = "localhost"
    port: int = 8080
    timeout: float = DEFAULT_TIMEOUT
    max_connections: int = MAX_POOL_SIZE
    protocol: str = "http"

    def endpoint_url(self: "ConnectionConfig") -> str:
        return self.protocol + "://" + self.host + ":" + str(self.port)

    def is_local(self: "ConnectionConfig") -> bool:
        return self.host == "localhost"

    def is_secure(self: "ConnectionConfig") -> bool:
        return self.protocol == "https"



DEFAULT_RETRY_POLICY: RetryPolicy = RetryPolicy()
DEFAULT_CONNECTION_CONFIG: ConnectionConfig = ConnectionConfig()
FATAL_ERROR: CoreError = CoreError()



import pytest

from blends.models import (
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.selection import (
    DefinitionCandidate,
    PartialPathEdge,
    edge_list_shadows,
    edge_shadows,
    prune_shadowed_candidates,
    select_definition_candidates_from_scope,
    sort_candidates_deterministically,
)
from blends.stack.view import (
    StackGraphView,
)


def _view_for_graph(graph: SyntaxGraph) -> StackGraphView:
    return StackGraphView.from_syntax_graph(graph, path="test.py")


def _add_scope_node(graph: SyntaxGraph, node_id: str) -> None:
    graph.add_node(node_id)
    graph.update_node_attribute(node_id, STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)


def _add_pop_symbol_node(
    graph: SyntaxGraph, node_id: str, symbol: str, *, is_definition: bool = True
) -> None:
    graph.add_node(node_id)
    graph.update_node_attribute(node_id, STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute(node_id, STACK_GRAPH_SYMBOL, symbol)
    graph.update_node_attribute(node_id, STACK_GRAPH_IS_DEFINITION, is_definition)


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("source1", "prec1", "source2", "prec2", "expected"),
    [
        (1, 2, 1, 1, True),
        (1, 1, 1, 1, False),
        (2, 2, 1, 1, False),
    ],
)
def test_edge_shadows(  # type: ignore[misc]
    source1: int,
    prec1: int,
    source2: int,
    prec2: int,
    expected: bool,  # noqa: FBT001
) -> None:
    assert (
        edge_shadows(
            PartialPathEdge(source_node_index=source1, precedence=prec1),
            PartialPathEdge(source_node_index=source2, precedence=prec2),
        )
        == expected
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_edge_list_shadows_is_not_comparable_when_sources_differ() -> None:
    assert not edge_list_shadows(
        (PartialPathEdge(source_node_index=1, precedence=2),),
        (PartialPathEdge(source_node_index=2, precedence=1),),
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_prune_shadowed_candidates_keeps_ties() -> None:
    candidates = [
        DefinitionCandidate(
            definition_node_index=10,
            edges=(PartialPathEdge(source_node_index=1, precedence=0),),
        ),
        DefinitionCandidate(
            definition_node_index=11,
            edges=(PartialPathEdge(source_node_index=1, precedence=0),),
        ),
    ]

    pruned = prune_shadowed_candidates(candidates)

    assert {c.definition_node_index for c in pruned} == {10, 11}


@pytest.mark.blends_test_group("stack_unittesting")
def test_prune_shadowed_candidates_removes_lower_precedence_for_same_source() -> None:
    candidates = [
        DefinitionCandidate(
            definition_node_index=10,
            edges=(PartialPathEdge(source_node_index=1, precedence=0),),
        ),
        DefinitionCandidate(
            definition_node_index=11,
            edges=(PartialPathEdge(source_node_index=1, precedence=1),),
        ),
    ]

    pruned = prune_shadowed_candidates(candidates)

    assert [c.definition_node_index for c in sort_candidates_deterministically(pruned)] == [11]


@pytest.mark.blends_test_group("stack_unittesting")
def test_sort_candidates_is_deterministic() -> None:
    a = DefinitionCandidate(
        definition_node_index=20,
        edges=(PartialPathEdge(source_node_index=1, precedence=0),),
    )
    b = DefinitionCandidate(
        definition_node_index=10,
        edges=(PartialPathEdge(source_node_index=1, precedence=0),),
    )

    assert sort_candidates_deterministically([a, b]) == [b, a]
    assert sort_candidates_deterministically([b, a]) == [b, a]


@pytest.mark.blends_test_group("stack_unittesting")
def test_select_definition_candidates_from_scope_prunes_by_edge_precedence() -> None:
    graph = SyntaxGraph()
    _add_scope_node(graph, "2")
    _add_pop_symbol_node(graph, "3", "x")
    _add_pop_symbol_node(graph, "4", "x")

    graph.add_edge("2", "3", precedence=0)
    graph.add_edge("2", "4", precedence=1)

    view = _view_for_graph(graph)
    selected = select_definition_candidates_from_scope(
        view, view.nid_to_index["2"], view.symbols.intern("x")
    )

    assert selected == [view.nid_to_index["4"]]


@pytest.mark.blends_test_group("stack_unittesting")
def test_view_outgoing_is_sorted_by_sink_index() -> None:
    graph = SyntaxGraph()
    _add_scope_node(graph, "2")
    _add_pop_symbol_node(graph, "3", "x", is_definition=False)
    _add_pop_symbol_node(graph, "4", "x", is_definition=False)

    add_edge(graph, Edge(source="2", sink="4", precedence=0))
    add_edge(graph, Edge(source="2", sink="3", precedence=0))

    view = _view_for_graph(graph)
    assert view.outgoing[view.nid_to_index["2"]] == [
        (view.nid_to_index["3"], 0),
        (view.nid_to_index["4"], 0),
    ]


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("first_precedence", "second_precedence", "expected_precedence"),
    [(0, 2, 2), (2, 0, 2)],
)
def test_add_edge_keeps_max_precedence_for_existing_edge(  # type: ignore[misc]
    first_precedence: int, second_precedence: int, expected_precedence: int
) -> None:
    graph = SyntaxGraph()
    _add_scope_node(graph, "2")
    _add_pop_symbol_node(graph, "3", "x", is_definition=False)

    add_edge(graph, Edge(source="2", sink="3", precedence=first_precedence))
    add_edge(graph, Edge(source="2", sink="3", precedence=second_precedence))

    view = _view_for_graph(graph)
    assert view.outgoing[view.nid_to_index["2"]] == [(view.nid_to_index["3"], expected_precedence)]



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    cp_node = graph.nodes[args.n_id]
    identifier_id = cp_node["label_field_name"]
    identifier_name = node_to_str(graph, identifier_id)
    variable_type = None
    if catch_type_id := match_ast_d(graph, args.n_id, "catch_type"):
        variable_type = node_to_str(graph, catch_type_id)

    return build_parameter_node(
        args=args,
        variable=identifier_name,
        variable_type=variable_type,
        value_id=identifier_id,
    )



import pytest

from test.stack.test_helpers import (
    find_symbol_nodes,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("functional")
def test_single_file_resolver_end_to_end_python() -> None:
    view, record = setup_stack_resolution("stack_graph_single_file.py")

    symbol_id = view.symbols.symbol_to_id.get("x")
    assert symbol_id is not None

    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)

    assert len(ref_nodes) == 1
    assert len(def_nodes) == 1

    result = resolve_symbol(view, record, "x")
    assert result == [def_nodes[0]]


@pytest.mark.blends_test_group("functional")
def test_single_file_resolver_end_to_end_python_scopes() -> None:
    view, record = setup_stack_resolution("stack_graph_single_file_scopes.py")

    result_x = resolve_symbol(view, record, "x")
    assert len(result_x) == 1, "x should resolve to one definition"

    result_y = resolve_symbol(view, record, "y")
    assert len(result_y) == 1, "y should resolve to one definition"


@pytest.mark.blends_test_group("functional")
def test_single_file_resolver_nested_scopes() -> None:
    view, record = setup_stack_resolution("stack_graph_single_file_nested.py")

    result_x = resolve_symbol(view, record, "x")
    assert len(result_x) == 1, "x should resolve to one definition"

    result_y = resolve_symbol(view, record, "y")
    assert len(result_y) == 1, "y should resolve to one definition"

    result_z = resolve_symbol(view, record, "z")
    assert len(result_z) == 1, "z should resolve to one definition"


@pytest.mark.blends_test_group("functional")
def test_single_file_resolver_shadowing() -> None:
    view, record = setup_stack_resolution("stack_graph_single_file_complex.py")

    symbol_id = view.symbols.symbol_to_id.get("x")
    assert symbol_id is not None

    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)

    assert len(ref_nodes) == 1
    assert len(def_nodes) == 2  # noqa: PLR2004

    result = resolve_symbol(view, record, "x")
    assert len(result) == 2, (  # noqa: PLR2004
        "Uniform precedence should return all same-scope defs of x"
    )

    result_y = resolve_symbol(view, record, "y")
    assert len(result_y) == 1, "y should resolve to one definition"



from blends.models import (
    NId,
)
from blends.syntax.builders.switch_statement import (
    build_switch_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    body_id = args.ast_graph.nodes[args.n_id]["label_field_body"]
    value_id = args.ast_graph.nodes[args.n_id]["label_field_value"]
    return build_switch_statement_node(args, body_id, value_id)



from collections.abc import (
    Iterator,
)

from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def method_modifies_symbol(args: SearchArgs) -> bool:
    n_attr = args.graph.nodes[args.n_id]
    expr_split = n_attr["expression"].split(".")
    obj_id = n_attr.get("object_id")
    return (obj_id and args.symbol == args.graph.nodes[obj_id].get("symbol")) or (
        len(expr_split) > 1 and args.symbol == expr_split[0]
    )


def search(args: SearchArgs) -> Iterator[SearchResult]:
    if not args.def_only and method_modifies_symbol(args):
        yield False, args.n_id



from blends.stack.criteria import (
    is_complete_definition_path,
    is_definition_endpoint,
    is_endpoint,
    is_exported_scope_endpoint,
    is_import_binding_endpoint,
    is_jump_to_boundary,
    is_reference_endpoint,
    is_root_endpoint,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.forward_stitcher import (
    ForwardStitcher,
    ForwardStitcherConfig,
    ForwardStitcherLoadRequests,
    ForwardStitcherPhaseResult,
    ForwardStitcherStats,
    ForwardStitcherStatsSnapshot,
)
from blends.stack.multi_file_resolver import (
    MultiFileResolutionRequest,
    MultiFileResolutionStats,
    resolve_definitions_multi_file,
)
from blends.stack.node_helpers import (
    drop_scopes_node_attributes,
    import_binding_node_attributes,
    jump_to_node_attributes,
    pop_scoped_symbol_node_attributes,
    pop_symbol_node_attributes,
    push_scoped_symbol_node_attributes,
    push_symbol_node_attributes,
    root_node_attributes,
    scope_node_attributes,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
    is_stack_graph_kind,
)
from blends.stack.partial_path import (
    FileHandle,
    PartialPath,
    PartialPathDatabase,
    PartialPathDb,
    PartialPathFileRecord,
    PartialPathId,
    PartialPathIndexRequest,
    PartialPathLimits,
    PartialPathResolutionError,
    PartialPathResolutionErrorCode,
    PartialScopedSymbol,
    PartialScopeStack,
    PartialScopeStackBindings,
    PartialSymbolStack,
    PartialSymbolStackBindings,
    PerFilePartialPaths,
    PerFilePartialPathStats,
    ScopeStackVariable,
    StackGraphViewMetadata,
    SymbolStackVariable,
    compute_minimal_partial_paths_in_file,
    index_partial_paths_for_file,
    remap_symbol_id,
)
from blends.stack.selection import (
    DefinitionCandidate,
    PartialPathEdge,
    edge_list_shadows,
    edge_shadows,
    prune_shadowed_candidates,
    select_definition_candidates_from_scope,
    sort_candidates_deterministically,
)
from blends.stack.single_file_resolver import (
    SingleFileResolutionRequest,
    SingleFileResolutionStats,
    resolve_definitions_single_file,
)
from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)
from blends.stack.transitions import (
    TransitionError,
    TransitionResult,
    apply_node,
)
from blends.stack.validation import (
    validate_stack_graph_graph,
    validate_stack_graph_node,
)
from blends.stack.view import (
    StackGraphView,
)

__all__ = [
    "DefinitionCandidate",
    "Edge",
    "FileHandle",
    "ForwardStitcher",
    "ForwardStitcherConfig",
    "ForwardStitcherLoadRequests",
    "ForwardStitcherPhaseResult",
    "ForwardStitcherStats",
    "ForwardStitcherStatsSnapshot",
    "MultiFileResolutionRequest",
    "MultiFileResolutionStats",
    "PartialPath",
    "PartialPathDatabase",
    "PartialPathDb",
    "PartialPathEdge",
    "PartialPathFileRecord",
    "PartialPathId",
    "PartialPathIndexRequest",
    "PartialPathLimits",
    "PartialPathResolutionError",
    "PartialPathResolutionErrorCode",
    "PartialScopeStack",
    "PartialScopeStackBindings",
    "PartialScopedSymbol",
    "PartialSymbolStack",
    "PartialSymbolStackBindings",
    "PerFilePartialPathStats",
    "PerFilePartialPaths",
    "ScopeStackNode",
    "ScopeStackVariable",
    "SingleFileResolutionRequest",
    "SingleFileResolutionStats",
    "StackGraphNodeKind",
    "StackGraphView",
    "StackGraphViewMetadata",
    "StackState",
    "SymbolStackNode",
    "SymbolStackVariable",
    "TransitionError",
    "TransitionResult",
    "add_edge",
    "apply_node",
    "compute_minimal_partial_paths_in_file",
    "drop_scopes_node_attributes",
    "edge_list_shadows",
    "edge_shadows",
    "import_binding_node_attributes",
    "index_partial_paths_for_file",
    "is_complete_definition_path",
    "is_definition_endpoint",
    "is_endpoint",
    "is_exported_scope_endpoint",
    "is_import_binding_endpoint",
    "is_jump_to_boundary",
    "is_reference_endpoint",
    "is_root_endpoint",
    "is_stack_graph_kind",
    "jump_to_node_attributes",
    "pop_scoped_symbol_node_attributes",
    "pop_symbol_node_attributes",
    "prune_shadowed_candidates",
    "push_scoped_symbol_node_attributes",
    "push_symbol_node_attributes",
    "remap_symbol_id",
    "resolve_definitions_multi_file",
    "resolve_definitions_single_file",
    "root_node_attributes",
    "scope_node_attributes",
    "select_definition_candidates_from_scope",
    "sort_candidates_deterministically",
    "validate_stack_graph_graph",
    "validate_stack_graph_node",
]



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.hcl import (
    attribute as hcl_attribute,
)
from blends.syntax.readers.hcl import (
    block as hcl_block,
)
from blends.syntax.readers.hcl import (
    config_file as hcl_config_file,
)
from blends.syntax.readers.hcl import (
    expression as hcl_expression,
)
from blends.syntax.readers.hcl import (
    function_arguments as hcl_function_arguments,
)
from blends.syntax.readers.hcl import (
    identifier as hcl_identifier,
)

HCL_DISPATCHERS: Dispatcher = {
    "attribute": hcl_attribute.reader,
    "object_elem": hcl_attribute.reader,
    "block": hcl_block.reader,
    "config_file": hcl_config_file.reader,
    "expression": hcl_expression.reader,
    "function_arguments": hcl_function_arguments.reader,
    "identifier": hcl_identifier.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group,
)
from blends.syntax.builders.switch_body import (
    build_switch_body_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    def_n_id: NId | None = None

    c_ids = match_ast_group(graph, args.n_id, "case_statement", "default_statement")

    case_ids = c_ids.get("case_statement", [])

    if def_n_ids := c_ids.get("default_statement"):
        def_n_id = def_n_ids[0]

    return build_switch_body_node(args, iter(case_ids), def_n_id)



import pytest

from blends.stack.node_kinds import (
    StackGraphNodeKind,
    is_stack_graph_kind,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_all_node_kinds_defined() -> None:
    expected_kinds = {
        "DropScopesNode",
        "JumpToNode",
        "PopScopedSymbolNode",
        "PopSymbolNode",
        "PushScopedSymbolNode",
        "PushSymbolNode",
        "RootNode",
        "ScopeNode",
    }
    actual_kinds = {kind.value for kind in StackGraphNodeKind}
    assert actual_kinds == expected_kinds


@pytest.mark.blends_test_group("stack_unittesting")
def test_node_kind_string_conversion() -> None:
    assert StackGraphNodeKind.PUSH_SYMBOL.value == "PushSymbolNode"
    assert StackGraphNodeKind.POP_SYMBOL.value == "PopSymbolNode"
    assert StackGraphNodeKind.SCOPE.value == "ScopeNode"
    assert StackGraphNodeKind.ROOT.value == "RootNode"


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_stack_graph_kind_valid_kinds() -> None:
    assert is_stack_graph_kind("PushSymbolNode") is True
    assert is_stack_graph_kind("PopSymbolNode") is True
    assert is_stack_graph_kind("ScopeNode") is True
    assert is_stack_graph_kind("RootNode") is True
    assert is_stack_graph_kind("JumpToNode") is True
    assert is_stack_graph_kind("DropScopesNode") is True
    assert is_stack_graph_kind("PushScopedSymbolNode") is True
    assert is_stack_graph_kind("PopScopedSymbolNode") is True


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_stack_graph_kind_invalid_strings() -> None:
    assert is_stack_graph_kind("InvalidNode") is False
    assert is_stack_graph_kind("push_symbol") is False
    assert is_stack_graph_kind("") is False
    assert is_stack_graph_kind("Scope") is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_is_stack_graph_kind_non_string_types() -> None:
    assert is_stack_graph_kind(123) is False
    assert is_stack_graph_kind(None) is False
    assert is_stack_graph_kind([]) is False
    assert is_stack_graph_kind({}) is False
    assert is_stack_graph_kind(StackGraphNodeKind.PUSH_SYMBOL) is True



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_name = "Unnamed"
    var_type = None

    var_dec = match_ast_d(graph, args.n_id, "variable_declaration")
    if var_dec:
        name = match_ast_d(graph, var_dec, "identifier")
        if name:
            var_name = node_to_str(graph, name)

        var_type_id = match_ast_d(graph, var_dec, "user_type")
        if var_type_id:
            var_type = node_to_str(graph, var_type_id)

    value_id = adj_ast(graph, args.n_id)[-1]
    if graph.nodes[value_id]["label_type"] in {
        "property_delegate",
        "variable_declaration",
    }:
        val_id = match_ast_d(graph, value_id, "call_expression")
        return build_variable_declaration_node(args, var_name, var_type, val_id)

    return build_variable_declaration_node(args, var_name, var_type, value_id)



import pytest

from blends.models import (
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)
from blends.stack.transitions import (
    TransitionError,
    apply_node,
)
from blends.stack.view import (
    JUMP_TO_INDEX,
    StackGraphView,
)


def _view_for_graph(graph: SyntaxGraph) -> StackGraphView:
    return StackGraphView.from_syntax_graph(graph, path="test.py")


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_push_symbol_pushes_symbol() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    state = StackState(symbol_stack=None, scope_stack=None)

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error is None
    assert result.jump_target is None
    assert result.state.symbol_stack is not None
    assert view.symbols.lookup(result.state.symbol_stack.symbol_id) == "x"
    assert result.state.symbol_stack.scopes is None
    assert result.state.symbol_stack.tail is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_pop_symbol_pops_only_when_symbol_matches() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    sym_id = view.symbols.intern("x")
    state = StackState(symbol_stack=SymbolStackNode(sym_id, None, None), scope_stack=None)

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error is None
    assert result.state.symbol_stack is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_pop_symbol_errors_on_empty_symbol_stack() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    state = StackState(symbol_stack=None, scope_stack=None)

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error == TransitionError.EMPTY_SYMBOL_STACK


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_pop_symbol_errors_on_symbol_mismatch() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    state = StackState(
        symbol_stack=SymbolStackNode(view.symbols.intern("y"), None, None), scope_stack=None
    )

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error == TransitionError.INCORRECT_POPPED_SYMBOL


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_pop_symbol_errors_on_unexpected_attached_scopes() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    attached = ScopeStackNode(scope_index=123, tail=None)
    state = StackState(
        symbol_stack=SymbolStackNode(view.symbols.intern("x"), attached, None),
        scope_stack=None,
    )

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error == TransitionError.UNEXPECTED_ATTACHED_SCOPE_LIST


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_drop_scopes_clears_scope_stack() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)

    view = _view_for_graph(graph)
    state = StackState(symbol_stack=None, scope_stack=ScopeStackNode(1, None))

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error is None
    assert result.state.scope_stack is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_jump_to_errors_on_empty_scope_stack() -> None:
    graph = SyntaxGraph()

    view = _view_for_graph(graph)
    state = StackState(symbol_stack=None, scope_stack=None)

    result = apply_node(view, JUMP_TO_INDEX, state)

    assert result.error == TransitionError.EMPTY_SCOPE_STACK


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_jump_to_returns_top_scope_and_pops_it() -> None:
    graph = SyntaxGraph()

    view = _view_for_graph(graph)
    scope_stack = ScopeStackNode(scope_index=10, tail=ScopeStackNode(scope_index=20, tail=None))
    state = StackState(symbol_stack=None, scope_stack=scope_stack)

    result = apply_node(view, JUMP_TO_INDEX, state)

    assert result.error is None
    assert result.jump_target == 10  # noqa: PLR2004
    assert result.state.scope_stack is not None
    assert result.state.scope_stack.scope_index == 20  # noqa: PLR2004
    assert result.state.scope_stack.tail is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_push_scoped_symbol_attaches_scope_and_current_scope_stack() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("3", STACK_GRAPH_SCOPE, "2")

    view = _view_for_graph(graph)
    existing_scope_stack = ScopeStackNode(scope_index=999, tail=None)
    state = StackState(symbol_stack=None, scope_stack=existing_scope_stack)

    result = apply_node(view, view.nid_to_index["3"], state)

    assert result.error is None
    assert result.state.symbol_stack is not None
    assert result.state.symbol_stack.scopes is not None
    assert result.state.symbol_stack.scopes.scope_index == view.nid_to_index["2"]
    assert result.state.symbol_stack.scopes.tail == existing_scope_stack
    assert result.state.scope_stack == existing_scope_stack


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_pop_scoped_symbol_restores_scope_stack_from_attached_scopes() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SCOPED_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    attached_scopes = ScopeStackNode(scope_index=7, tail=ScopeStackNode(scope_index=8, tail=None))
    top = SymbolStackNode(symbol_id=view.symbols.intern("x"), scopes=attached_scopes, tail=None)
    state = StackState(symbol_stack=top, scope_stack=ScopeStackNode(scope_index=1, tail=None))

    result = apply_node(view, view.nid_to_index["2"], state)

    assert result.error is None
    assert result.state.symbol_stack is None
    assert result.state.scope_stack == attached_scopes


@pytest.mark.blends_test_group("stack_unittesting")
def test_apply_node_ref_to_def_traversal_ends_with_empty_stacks() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("2", STACK_GRAPH_SYMBOL, "x")

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("3", STACK_GRAPH_SYMBOL, "x")

    view = _view_for_graph(graph)
    state = StackState(symbol_stack=None, scope_stack=None)

    result1 = apply_node(view, view.nid_to_index["2"], state)
    assert result1.error is None
    assert result1.state.symbol_stack is not None

    result2 = apply_node(view, view.nid_to_index["3"], result1.state)
    assert result2.error is None
    assert result2.state.symbol_stack is None
    assert result2.state.scope_stack is None



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.execution_block import (
    build_execution_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    _, *c_ids, _ = adj_ast(graph, args.n_id)
    ignored_labels = {
        "\n",
        "\r\n",
    }

    filtered_ids: set[NId] = set()

    for c_id in c_ids:
        label_type = graph.nodes[c_id]["label_type"]
        if label_type == "expression_statement":
            filtered_ids.update(adj_ast(graph, c_id))
        elif label_type in ignored_labels:
            continue
        else:
            filtered_ids.add(c_id)

    return build_execution_block_node(args, iter(filtered_ids))



from collections.abc import Iterable

from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_symbol_node_attributes,
    scope_node_attributes,
)
from blends.stack.policies.registry import (
    get_export_policy,
)
from blends.stack.scope_types import ScopeType
from blends.syntax.builders.module_path import compute_module_path_segments
from blends.syntax.builders.utils import (
    get_next_synthetic_node_id,
    get_or_create_root_nid,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _build_module_definition_path(args: SyntaxGraphArgs, file_scope_id: NId) -> None:
    segments = compute_module_path_segments(args.path)
    if not segments:
        return

    root_nid = get_or_create_root_nid(args)
    previous_id: NId = root_nid
    for index, segment in enumerate(segments):
        segment_id = get_next_synthetic_node_id(args)
        args.syntax_graph.add_node(
            segment_id,
            label_type="SyntheticModulePath",
            **pop_symbol_node_attributes(symbol=segment, is_definition=False),
        )
        add_edge(args.syntax_graph, Edge(source=previous_id, sink=segment_id))

        if index < len(segments) - 1:
            dot_id = get_next_synthetic_node_id(args)
            args.syntax_graph.add_node(
                dot_id,
                label_type="SyntheticModulePath",
                **pop_symbol_node_attributes(symbol=".", is_definition=False),
            )
            add_edge(args.syntax_graph, Edge(source=segment_id, sink=dot_id))
            previous_id = dot_id
        else:
            previous_id = segment_id

    add_edge(args.syntax_graph, Edge(source=previous_id, sink=file_scope_id))


def build_file_node(args: SyntaxGraphArgs, c_ids: Iterable[NId]) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="File",
    )

    if ctx.has_feature_flag("StackGraph"):
        policy = get_export_policy(args.language)
        args.syntax_graph.update_node(
            args.n_id,
            scope_node_attributes(
                is_exported=policy.is_scope_exported(
                    args.syntax_graph, args.n_id, ScopeType.FILE.value
                ),
                scope_type=ScopeType.FILE,
            ),
        )
        _build_module_definition_path(args, file_scope_id=args.n_id)

    for c_id in c_ids:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(c_id)),
            label_ast="AST",
        )

    return args.n_id



from __future__ import annotations

from mf_core import (
    SENTINEL,
    MAX_POOL_SIZE,
    DEFAULT_TIMEOUT,
    HTTP_OK,
    HTTP_UNAVAILABLE,
    CoreError,
    RetryPolicy,
    ConnectionConfig,
)


class TransportError(CoreError):
    transport_code: int = 0

    def describe(self: "TransportError") -> str:
        base: str = CoreError.describe(self)
        return base + " transport=" + str(self.transport_code)

    def is_retryable(self: "TransportError") -> bool:
        return self.transport_code < HTTP_UNAVAILABLE


class ConnectionPool:
    config: ConnectionConfig = ConnectionConfig()
    active: int = 0
    peak: int = 0
    failed_acquires: int = 0
    total_acquired: int = 0

    def acquire(self: "ConnectionPool") -> bool:
        limit: int = MAX_POOL_SIZE
        if self.active >= limit:
            self.failed_acquires = self.failed_acquires + 1
            return False
        self.active = self.active + 1
        self.total_acquired = self.total_acquired + 1
        if self.active > self.peak:
            self.peak = self.active
        return True

    def release(self: "ConnectionPool") -> None:
        if self.active > 0:
            self.active = self.active - 1

    def utilization(self: "ConnectionPool") -> int:
        limit: int = MAX_POOL_SIZE
        if limit == 0:
            return 0
        return self.active * 100 // limit

    def is_saturated(self: "ConnectionPool") -> bool:
        return self.active >= MAX_POOL_SIZE


class Request:
    path: str = "/"
    payload_size: int = 0
    timeout: float = DEFAULT_TIMEOUT
    retry: RetryPolicy = RetryPolicy()
    attempt: int = 0

    def can_retry(self: "Request") -> bool:
        return self.retry.should_retry(self.attempt)

    def next_delay(self: "Request") -> int:
        return self.retry.delay_ms(self.attempt)

    def is_idempotent(self: "Request") -> bool:
        return self.attempt == 0


class Response:
    status: int = 0
    payload_size: int = 0
    latency_ms: int = 0
    error: TransportError = TransportError()

    def is_success(self: "Response") -> bool:
        return HTTP_OK <= self.status < 300

    def is_error(self: "Response") -> bool:
        return self.status >= 400

    def is_unavailable(self: "Response") -> bool:
        return self.status == HTTP_UNAVAILABLE


def send_request(pool: ConnectionPool, req: Request) -> Response:
    resp: Response = Response()
    if not pool.acquire():
        resp.status = HTTP_UNAVAILABLE
        resp.error.code = HTTP_UNAVAILABLE
        return resp
    resp.status = HTTP_OK
    resp.latency_ms = req.next_delay()
    pool.release()
    return resp


def send_with_retry(pool: ConnectionPool, req: Request) -> Response:
    resp: Response = Response()
    while req.can_retry():
        resp = send_request(pool, req)
        if resp.is_success():
            return resp
        req.attempt = req.attempt + 1
    sentinel_resp: Response = Response()
    sentinel_resp.status = SENTINEL
    sentinel_resp.error.code = SENTINEL
    return sentinel_resp


GLOBAL_POOL: ConnectionPool = ConnectionPool()
GLOBAL_TIMEOUT: float = DEFAULT_TIMEOUT
DEFAULT_RETRY: RetryPolicy = RetryPolicy()



from collections.abc import (
    Iterator,
)

from blends.query import (
    adj_ast,
)
from blends.symbolic_evaluation.context.search.model import (
    SearchArgs,
    SearchResult,
)


def search(args: SearchArgs) -> Iterator[SearchResult]:
    init_id = args.graph.nodes[args.n_id].get("initializer_id")
    if init_id:
        init_node = args.graph.nodes[init_id]
        if init_node.get("label_type") == "SymbolLookup" and args.symbol == init_node.get("symbol"):
            yield True, args.n_id
        else:
            symbols_ids = adj_ast(args.graph, init_id, label_type="SymbolLookup")
            for c_id in symbols_ids:
                if args.symbol == args.graph.nodes[c_id].get("symbol"):
                    yield True, args.n_id
                    break



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SCOPE_TYPE,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.view import (
    JUMP_TO_INDEX,
    ROOT_INDEX,
    Degree,
    StackGraphView,
    SymbolInterner,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_ignores_nodes_with_none_kind() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_with_none_kind")
    graph.update_node_attribute("node_with_none_kind", STACK_GRAPH_KIND, None)

    graph.add_node("node_without_kind")

    graph.add_node("valid_node")
    graph.update_node_attribute("valid_node", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    view = StackGraphView.from_syntax_graph(graph)

    assert "valid_node" in view.nid_to_index
    assert "node_with_none_kind" not in view.nid_to_index
    assert "node_without_kind" not in view.nid_to_index


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_singleton_indices_are_fixed() -> None:
    graph = SyntaxGraph()

    view = StackGraphView.from_syntax_graph(graph)

    assert view.node_kind[ROOT_INDEX] == StackGraphNodeKind.ROOT.value
    assert view.node_kind[JUMP_TO_INDEX] == StackGraphNodeKind.JUMP_TO.value
    assert len(view.index_to_nid) >= 2  # noqa: PLR2004


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_sorts_numeric_node_ids_by_int_value() -> None:
    graph = SyntaxGraph()

    graph.add_node("10")
    graph.update_node_attribute("10", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("a")
    graph.update_node_attribute("a", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    view = StackGraphView.from_syntax_graph(graph)

    assert view.nid_to_index["2"] < view.nid_to_index["10"]
    assert view.nid_to_index["10"] < view.nid_to_index["a"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_includes_edges_only_between_stack_nodes() -> None:
    graph = SyntaxGraph()

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("non_stack")

    graph.add_edge("2", "3", precedence=0)
    graph.add_edge("2", "non_stack", precedence=0)
    graph.add_edge("non_stack", "3", precedence=0)

    view = StackGraphView.from_syntax_graph(graph)

    source_index = view.nid_to_index["2"]
    sink_index = view.nid_to_index["3"]

    assert view.outgoing[source_index] == [(sink_index, 0)]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_sorts_outgoing_edges_by_sink_index() -> None:
    graph = SyntaxGraph()

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("4")
    graph.update_node_attribute("4", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_edge("2", "4", precedence=0)
    graph.add_edge("2", "3", precedence=0)

    view = StackGraphView.from_syntax_graph(graph)

    source_index = view.nid_to_index["2"]
    first_sink_index = view.nid_to_index["3"]
    second_sink_index = view.nid_to_index["4"]

    assert view.outgoing[source_index] == [
        (first_sink_index, 0),
        (second_sink_index, 0),
    ]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_computes_incoming_degree_summary() -> None:
    graph = SyntaxGraph()

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_node("4")
    graph.update_node_attribute("4", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    graph.add_edge("2", "3", precedence=0)
    graph.add_edge("4", "3", precedence=0)
    graph.add_edge("2", "4", precedence=0)

    view = StackGraphView.from_syntax_graph(graph)

    first_source_index = view.nid_to_index["2"]
    shared_sink_index = view.nid_to_index["3"]
    second_source_index = view.nid_to_index["4"]

    assert view.incoming_degree[first_source_index] == Degree.ZERO
    assert view.incoming_degree[second_source_index] == Degree.ONE
    assert view.incoming_degree[shared_sink_index] == Degree.MULTIPLE


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_collects_exported_scopes() -> None:
    graph = SyntaxGraph()

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=True)

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("3", STACK_GRAPH_IS_EXPORTED, value=False)

    view = StackGraphView.from_syntax_graph(graph)

    assert view.nid_to_index["2"] in view.exported_scopes
    assert view.nid_to_index["3"] not in view.exported_scopes


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_uses_path_argument_and_builds_file_maps() -> None:
    graph = SyntaxGraph()
    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    view = StackGraphView.from_syntax_graph(graph, path="override.py")

    assert view.file_path == "override.py"
    assert "override.py" in view.file_to_nodes
    assert view.nid_to_index["2"] in view.file_to_nodes["override.py"]
    assert view.node_to_file[view.nid_to_index["2"]] == "override.py"
    assert view.node_to_file[JUMP_TO_INDEX] == ""


@pytest.mark.blends_test_group("stack_unittesting")
def test_symbol_interner_reuses_cached_symbol_id_zero() -> None:
    interner = SymbolInterner(symbol_to_id={}, id_to_symbol=[])

    first_id = interner.intern("x")
    second_id = interner.intern("x")

    assert first_id == 0
    assert second_id == 0
    assert interner.symbol_to_id["x"] == 0
    assert interner.id_to_symbol == ["x"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_resolves_path_from_metadata_node_zero() -> None:
    graph = SyntaxGraph()
    graph.add_node("0")
    graph.update_node_attribute("0", "path", "from_metadata.py")

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)

    view = StackGraphView.from_syntax_graph(graph)

    assert view.file_path == "from_metadata.py"
    assert "from_metadata.py" in view.file_to_nodes


@pytest.mark.blends_test_group("stack_unittesting")
def test_stack_graph_view_extracts_node_scope_type() -> None:
    graph = SyntaxGraph()

    graph.add_node("2")
    graph.update_node_attribute("2", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("2", STACK_GRAPH_IS_EXPORTED, value=True)
    graph.update_node_attribute("2", STACK_GRAPH_SCOPE_TYPE, "file")

    graph.add_node("3")
    graph.update_node_attribute("3", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("3", STACK_GRAPH_IS_EXPORTED, value=False)
    graph.update_node_attribute("3", STACK_GRAPH_SCOPE_TYPE, "class")

    graph.add_node("4")
    graph.update_node_attribute("4", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("4", STACK_GRAPH_IS_EXPORTED, value=False)

    view = StackGraphView.from_syntax_graph(graph)

    assert view.node_scope_type[view.nid_to_index["2"]] == "file"
    assert view.node_scope_type[view.nid_to_index["3"]] == "class"
    assert view.node_scope_type[view.nid_to_index["4"]] is None



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    child_ids = adj_ast(graph, args.n_id)
    condition_id = child_ids[2]
    if (
        graph.nodes[condition_id]["label_type"] == "parenthesized_expression"
        and (childs := adj_ast(graph, condition_id))
        and len(childs) > 1
    ):
        condition_id = childs[1]

    true_id = child_ids[4]
    false_id = None
    expected_children_for_if_with_else = 7
    if len(child_ids) == expected_children_for_if_with_else:
        false_id = child_ids[6]
    return build_if_node(args, condition_id, true_id, false_id)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.ruby import (
    argument_list as ruby_argument_list,
)
from blends.syntax.readers.ruby import (
    array_node as ruby_array_node,
)
from blends.syntax.readers.ruby import (
    assignment as ruby_assignment,
)
from blends.syntax.readers.ruby import (
    begin as ruby_begin,
)
from blends.syntax.readers.ruby import (
    binary_expression as ruby_binary_expression,
)
from blends.syntax.readers.ruby import (
    boolean_literal as ruby_boolean_literal,
)
from blends.syntax.readers.ruby import (
    call as ruby_call,
)
from blends.syntax.readers.ruby import (
    catch_declaration as ruby_catch_declaration,
)
from blends.syntax.readers.ruby import (
    class_definition as ruby_class_definition,
)
from blends.syntax.readers.ruby import (
    comment as ruby_comment,
)
from blends.syntax.readers.ruby import (
    element_reference as ruby_element_reference,
)
from blends.syntax.readers.ruby import (
    exception_variable as ruby_exception_variable,
)
from blends.syntax.readers.ruby import (
    execution_block as ruby_execution_block,
)
from blends.syntax.readers.ruby import (
    for_statement as ruby_for_statement,
)
from blends.syntax.readers.ruby import (
    hash as ruby_hash,
)
from blends.syntax.readers.ruby import (
    identifier as ruby_identifier,
)
from blends.syntax.readers.ruby import (
    if_conditional as ruby_if_conditional,
)
from blends.syntax.readers.ruby import (
    if_modifier as ruby_if_modifier,
)
from blends.syntax.readers.ruby import (
    integer as ruby_integer,
)
from blends.syntax.readers.ruby import (
    method_declaration as ruby_method_declaration,
)
from blends.syntax.readers.ruby import (
    method_parameters as ruby_method_parameters,
)
from blends.syntax.readers.ruby import (
    module_definition as ruby_module_definition,
)
from blends.syntax.readers.ruby import (
    pair as ruby_pair,
)
from blends.syntax.readers.ruby import (
    parameter as ruby_parameter,
)
from blends.syntax.readers.ruby import (
    parameters as ruby_parameters,
)
from blends.syntax.readers.ruby import (
    program as ruby_program,
)
from blends.syntax.readers.ruby import (
    return_statement as ruby_return_statement,
)
from blends.syntax.readers.ruby import (
    scope_resolution as ruby_scope_resolution,
)
from blends.syntax.readers.ruby import (
    string_content as ruby_string_content,
)
from blends.syntax.readers.ruby import (
    string_literal as ruby_string_literal,
)
from blends.syntax.readers.ruby import (
    subshell as ruby_subshell,
)
from blends.syntax.readers.ruby import (
    try_statement as ruby_try_statement,
)
from blends.syntax.readers.ruby import (
    unless as ruby_unless,
)
from blends.syntax.readers.ruby import (
    while_statement as ruby_while_statement,
)

RUBY_DISPATCHERS: Dispatcher = {
    "array": ruby_array_node.reader,
    "argument_list": ruby_argument_list.reader,
    "assignment": ruby_assignment.reader,
    "exception_variable": ruby_exception_variable.reader,
    "binary": ruby_binary_expression.reader,
    "block_parameter": ruby_parameter.reader,
    "block_parameters": ruby_parameters.reader,
    "body_statement": ruby_execution_block.reader,
    "call": ruby_call.reader,
    "rescue": ruby_catch_declaration.reader,
    "class": ruby_class_definition.reader,
    "comment": ruby_comment.reader,
    "constant": ruby_identifier.reader,
    "do": ruby_execution_block.reader,
    "do_block": ruby_execution_block.reader,
    "else": ruby_execution_block.reader,
    "elsif": ruby_if_conditional.reader,
    "false": ruby_boolean_literal.reader,
    "for": ruby_for_statement.reader,
    "hash": ruby_hash.reader,
    "hash_key_symbol": ruby_identifier.reader,
    "hash_splat_parameter": ruby_parameter.reader,
    "identifier": ruby_identifier.reader,
    "if_modifier": ruby_if_modifier.reader,
    "if": ruby_if_conditional.reader,
    "instance_variable": ruby_identifier.reader,
    "integer": ruby_integer.reader,
    "method": ruby_method_declaration.reader,
    "method_parameters": ruby_method_parameters.reader,
    "module": ruby_module_definition.reader,
    "optional_parameter": ruby_parameter.reader,
    "pair": ruby_pair.reader,
    "program": ruby_program.reader,
    "regex": ruby_string_literal.reader,
    "return": ruby_return_statement.reader,
    "scope_resolution": ruby_scope_resolution.reader,
    "simple_symbol": ruby_identifier.reader,
    "singleton_class": ruby_class_definition.reader,
    "singleton_method": ruby_method_declaration.reader,
    "splat_parameter": ruby_parameter.reader,
    "string": ruby_string_literal.reader,
    "string_content": ruby_string_content.reader,
    "subshell": ruby_subshell.reader,
    "then": ruby_execution_block.reader,
    "true": ruby_boolean_literal.reader,
    "while": ruby_while_statement.reader,
    "while_modifier": ruby_while_statement.reader,
    "block": ruby_execution_block.reader,
    "block_body": ruby_execution_block.reader,
    "rescue_modifier": ruby_try_statement.reader,
    "begin": ruby_begin.reader,
    "element_reference": ruby_element_reference.reader,
    "unless": ruby_unless.reader,
    "unless_modifier": ruby_unless.reader,
}



from blends.models import (
    NId,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    body_id = n_attrs["label_field_body"]

    direct: DirectChildren = {
        "block_id": body_id,
        "parameters_id": n_attrs["label_field_parameters"],
    }

    return build_method_declaration_node(args, "LambdaExpression", direct, {})
